import json

from twisted.internet.defer import inlineCallbacks

from . import wormhole
from .tor_manager import get_tor


@inlineCallbacks
def receive(reactor,
            appid,
            relay_url,
            code,
            use_tor=False,
            launch_tor=False,
            tor_control_port=None,
            on_code=None):
    """
    This is a convenience API which returns a Deferred that callbacks
    with a single chunk of data from another wormhole (and then closes
    the wormhole). Under the hood, it's just using an instance
    returned from :func:`wormhole.wormhole`. This is similar to the
    `wormhole receive` command.

    :param unicode appid: our application ID

    :param unicode relay_url: the relay URL to use

    :param unicode code: a pre-existing code to use, or None

    :param bool use_tor: True if we should use Tor, False to not use it (None
                         for default)

    :param on_code: if not None, this is called when we have a code (even if
                    you passed in one explicitly)
    :type on_code: single-argument callable
    """
    tor = None
    if use_tor:
        tor = yield get_tor(reactor, launch_tor, tor_control_port)
        # For now, block everything until Tor has started. Soon: launch
        # tor in parallel with everything else, make sure the Tor object
        # can lazy-provide an endpoint, and overlap the startup process
        # with the user handing off the wormhole code

    wh = wormhole.create(appid, relay_url, reactor, tor=tor)
    if code is None:
        wh.allocate_code()
        code = yield wh.get_code()
    else:
        wh.set_code(code)
    # we'll call this no matter what, even if you passed in a code --
    # maybe it should be only in the 'if' block above?
    if on_code:
        on_code(code)
    data = yield wh.get_message()
    data = json.loads(data.decode("utf-8"))
    offer = data.get('offer', None)
    if not offer:
        raise Exception(f"Do not understand response: {data}")
    msg = None
    if 'message' in offer:
        msg = offer['message']
        wh.send_message(
            json.dumps({
                "answer": {
                    "message_ack": "ok"
                }
            }).encode("utf-8"))

    else:
        raise Exception(f"Unknown offer type: {offer.keys()}")

    yield wh.close()
    return msg


@inlineCallbacks
def send(reactor,
         appid,
         relay_url,
         data,
         code,
         use_tor=False,
         launch_tor=False,
         tor_control_port=None,
         on_code=None):
    """
    This is a convenience API which returns a Deferred that callbacks
    after a single chunk of data has been sent to another
    wormhole. Under the hood, it's just using an instance returned
    from :func:`wormhole.wormhole`. This is similar to the `wormhole
    send` command.

    :param unicode appid: the application ID

    :param unicode relay_url: the relay URL to use

    :param unicode code: a pre-existing code to use, or None

    :param bool use_tor: True if we should use Tor, False to not use it (None
                         for default)

    :param on_code: if not None, this is called when we have a code (even if
                    you passed in one explicitly)

    :type on_code: single-argument callable
    """
    tor = None
    if use_tor:
        tor = yield get_tor(reactor, launch_tor, tor_control_port)
        # For now, block everything until Tor has started. Soon: launch
        # tor in parallel with everything else, make sure the Tor object
        # can lazy-provide an endpoint, and overlap the startup process
        # with the user handing off the wormhole code
    wh = wormhole.create(appid, relay_url, reactor, tor=tor)
    if code is None:
        wh.allocate_code()
        code = yield wh.get_code()
    else:
        wh.set_code(code)
    if on_code:
        on_code(code)

    wh.send_message(json.dumps({"offer": {"message": data}}).encode("utf-8"))
    data = yield wh.get_message()
    data = json.loads(data.decode("utf-8"))
    answer = data.get('answer', None)
    yield wh.close()
    if answer:
        return None
    else:
        raise Exception(f"Unknown answer: {data}")



from unittest import mock
from zope.interface import alsoProvides
from twisted.internet.interfaces import ITransport
from ..._dilation._noise import NoiseInvalidMessage, NoiseConnection
from ..._dilation.connection import (IFramer, Frame, Prologue,
                                     _Record, Handshake, KCM,
                                     Disconnect, Ping, _Framer, Data)
from ..._dilation.connector import build_noise
from ..._dilation.roles import LEADER, FOLLOWER
from zope.interface import implementer
import pytest


def make_record():
    f = mock.Mock()
    alsoProvides(f, IFramer)
    n = mock.Mock()  # pretends to be a Noise object
    r = _Record(f, n, LEADER)
    r.set_role_leader()
    return r, f, n


def test_good2():
    f = mock.Mock()
    alsoProvides(f, IFramer)
    f.add_and_parse = mock.Mock(side_effect=[
        [],
        [Prologue()],
        [Frame(frame=b"rx-handshake")],
        [Frame(frame=b"frame1"), Frame(frame=b"frame2")],
    ])
    n = mock.Mock()
    n.write_message = mock.Mock(return_value=b"tx-handshake")
    p1, p2 = object(), object()
    n.decrypt = mock.Mock(side_effect=[p1, p2])
    r = _Record(f, n, LEADER)
    r.set_role_leader()
    assert f.mock_calls == []
    r.connectionMade()
    assert f.mock_calls == [mock.call.connectionMade()]
    f.mock_calls[:] = []
    assert n.mock_calls == [mock.call.start_handshake()]
    n.mock_calls[:] = []

    # Pretend to deliver the prologue in two parts. The text we send in
    # doesn't matter: the side_effect= is what causes the prologue to be
    # recognized by the second call.
    assert list(r.add_and_unframe(b"pro")) == []
    assert f.mock_calls == [mock.call.add_and_parse(b"pro")]
    f.mock_calls[:] = []
    assert n.mock_calls == []

    # recognizing the prologue causes a handshake frame to be sent
    assert list(r.add_and_unframe(b"logue")) == []
    assert f.mock_calls == [mock.call.add_and_parse(b"logue"),
                                    mock.call.send_frame(b"tx-handshake")]
    f.mock_calls[:] = []
    assert n.mock_calls == [mock.call.write_message()]
    n.mock_calls[:] = []

    # next add_and_unframe is recognized as the Handshake
    assert list(r.add_and_unframe(b"blah")) == [Handshake()]
    assert f.mock_calls == [mock.call.add_and_parse(b"blah")]
    f.mock_calls[:] = []
    assert n.mock_calls == [mock.call.read_message(b"rx-handshake")]
    n.mock_calls[:] = []

    # next is a pair of Records
    r1, r2 = object(), object()
    with mock.patch("wormhole._dilation.connection.parse_record",
                    side_effect=[r1, r2]) as pr:
        assert list(r.add_and_unframe(b"blah2")) == [r1, r2]
        assert n.mock_calls == [mock.call.decrypt(b"frame1"),
                                        mock.call.decrypt(b"frame2")]
        assert pr.mock_calls == [mock.call(p1), mock.call(p2)]

def test_bad_handshake():
    f = mock.Mock()
    alsoProvides(f, IFramer)
    f.add_and_parse = mock.Mock(return_value=[Prologue(),
                                              Frame(frame=b"rx-handshake")])
    n = mock.Mock()
    n.write_message = mock.Mock(return_value=b"tx-handshake")
    nvm = NoiseInvalidMessage()
    n.read_message = mock.Mock(side_effect=nvm)
    r = _Record(f, n, LEADER)
    r.set_role_leader()
    assert f.mock_calls == []
    r.connectionMade()
    assert f.mock_calls == [mock.call.connectionMade()]
    f.mock_calls[:] = []
    assert n.mock_calls == [mock.call.start_handshake()]
    n.mock_calls[:] = []

    with mock.patch("wormhole._dilation.connection.log.err") as le:
        with pytest.raises(Disconnect):
            list(r.add_and_unframe(b"data"))
    assert le.mock_calls == \
                     [mock.call(nvm, "bad inbound noise handshake")]

def test_bad_message():
    f = mock.Mock()
    alsoProvides(f, IFramer)
    f.add_and_parse = mock.Mock(return_value=[Prologue(),
                                              Frame(frame=b"rx-handshake"),
                                              Frame(frame=b"bad-message")])
    n = mock.Mock()
    n.write_message = mock.Mock(return_value=b"tx-handshake")
    nvm = NoiseInvalidMessage()
    n.decrypt = mock.Mock(side_effect=nvm)
    r = _Record(f, n, LEADER)
    r.set_role_leader()
    assert f.mock_calls == []
    r.connectionMade()
    assert f.mock_calls == [mock.call.connectionMade()]
    f.mock_calls[:] = []
    assert n.mock_calls == [mock.call.start_handshake()]
    n.mock_calls[:] = []

    with mock.patch("wormhole._dilation.connection.log.err") as le:
        with pytest.raises(Disconnect):
            list(r.add_and_unframe(b"data"))
    assert le.mock_calls == \
                     [mock.call(nvm, "bad inbound noise frame")]

def test_send_record():
    f = mock.Mock()
    alsoProvides(f, IFramer)
    n = mock.Mock()
    f1 = b"some bytes"
    n.encrypt = mock.Mock(return_value=f1)
    r1 = Ping(b"pingid")
    r = _Record(f, n, LEADER)
    r.set_role_leader()
    assert f.mock_calls == []
    m1 = b"some other bytes"
    with mock.patch("wormhole._dilation.connection.encode_record",
                    return_value=m1) as er:
        r.send_record(r1)
    assert er.mock_calls == [mock.call(r1)]
    assert n.mock_calls == [mock.call.start_handshake(),
                                    mock.call.encrypt(m1)]
    assert f.mock_calls == [mock.call.send_frame(f1)]

def test_good():
    # Exercise the success path. The Record instance is given each chunk
    # of data as it arrives on Protocol.dataReceived, and is supposed to
    # return a series of Tokens (maybe none, if the chunk was incomplete,
    # or more than one, if the chunk was larger). Internally, it delivers
    # the chunks to the Framer for unframing (which returns 0 or more
    # frames), manages the Noise decryption object, and parses any
    # decrypted messages into tokens (some of which are consumed
    # internally, others for delivery upstairs).
    #
    # in the normal flow, we get:
    #
    # |   | Inbound   | NoiseAction   | Outbound  | ToUpstairs |
    # |   | -         | -             | -         | -          |
    # | 1 |           |               | prologue  |            |
    # | 2 | prologue  |               |           |            |
    # | 3 |           | write_message | handshake |            |
    # | 4 | handshake | read_message  |           | Handshake  |
    # | 5 |           | encrypt       | KCM       |            |
    # | 6 | KCM       | decrypt       |           | KCM        |
    # | 7 | msg1      | decrypt       |           | msg1       |

    # 1: instantiating the Record instance causes the outbound prologue
    # to be sent

    # 2+3: receipt of the inbound prologue triggers creation of the
    # ephemeral key (the "handshake") by calling noise.write_message()
    # and then writes the handshake to the outbound transport

    # 4: when the peer's handshake is received, it is delivered to
    # noise.read_message(), which generates the shared key (enabling
    # noise.send() and noise.decrypt()). It also delivers the Handshake
    # token upstairs, which might (on the Follower) trigger immediate
    # transmission of the Key Confirmation Message (KCM)

    # 5: the outbound KCM is framed and fed into noise.encrypt(), then
    # sent outbound

    # 6: the peer's KCM is decrypted then delivered upstairs. The
    # Follower treats this as a signal that it should use this connection
    # (and drop all others).

    # 7: the peer's first message is decrypted, parsed, and delivered
    # upstairs. This might be an Open or a Data, depending upon what
    # queued messages were left over from the previous connection

    r, f, n = make_record()
    outbound_handshake = object()
    kcm, msg1 = object(), object()
    f_kcm, f_msg1 = object(), object()
    n.write_message = mock.Mock(return_value=outbound_handshake)
    n.decrypt = mock.Mock(side_effect=[kcm, msg1])
    n.encrypt = mock.Mock(side_effect=[f_kcm, f_msg1])
    f.add_and_parse = mock.Mock(side_effect=[[],  # no tokens yet
                                             [Prologue()],
                                             [Frame("f_handshake")],
                                             [Frame("f_kcm"),
                                              Frame("f_msg1")],
                                             ])

    assert f.mock_calls == []
    assert n.mock_calls == [mock.call.start_handshake()]
    n.mock_calls[:] = []

    # 1. The Framer is responsible for sending the prologue, so we don't
    # have to check that here, we just check that the Framer was told
    # about connectionMade properly.
    r.connectionMade()
    assert f.mock_calls == [mock.call.connectionMade()]
    assert n.mock_calls == []
    f.mock_calls[:] = []

    # 2
    # we dribble the prologue in over two messages, to make sure we can
    # handle a dataReceived that doesn't complete the token

    # remember, add_and_unframe is a generator
    assert list(r.add_and_unframe(b"pro")) == []
    assert f.mock_calls == [mock.call.add_and_parse(b"pro")]
    assert n.mock_calls == []
    f.mock_calls[:] = []

    assert list(r.add_and_unframe(b"logue")) == []
    # 3: write_message, send outbound handshake
    assert f.mock_calls == [mock.call.add_and_parse(b"logue"),
                                    mock.call.send_frame(outbound_handshake),
                                    ]
    assert n.mock_calls == [mock.call.write_message()]
    f.mock_calls[:] = []
    n.mock_calls[:] = []

    # 4
    # Now deliver the Noise "handshake", the ephemeral public key. This
    # is framed, but not a record, so it shouldn't decrypt or parse
    # anything, but the handshake is delivered to the Noise object, and
    # it does return a Handshake token so we can let the next layer up
    # react (by sending the KCM frame if we're a Follower, or not if
    # we're the Leader)

    assert list(r.add_and_unframe(b"handshake")) == [Handshake()]
    assert f.mock_calls == [mock.call.add_and_parse(b"handshake")]
    assert n.mock_calls == [mock.call.read_message("f_handshake")]
    f.mock_calls[:] = []
    n.mock_calls[:] = []

    # 5: at this point we ought to be able to send a message, the KCM
    with mock.patch("wormhole._dilation.connection.encode_record",
                    side_effect=[b"r-kcm"]) as er:
        r.send_record(kcm)
    assert er.mock_calls == [mock.call(kcm)]
    assert n.mock_calls == [mock.call.encrypt(b"r-kcm")]
    assert f.mock_calls == [mock.call.send_frame(f_kcm)]
    n.mock_calls[:] = []
    f.mock_calls[:] = []

    # 6: Now we deliver two messages stacked up: the KCM (Key
    # Confirmation Message) and the first real message. Concatenating
    # them tests that we can handle more than one token in a single
    # chunk. We need to mock parse_record() because everything past the
    # handshake is decrypted and parsed.

    with mock.patch("wormhole._dilation.connection.parse_record",
                    side_effect=[kcm, msg1]) as pr:
        assert list(r.add_and_unframe(b"kcm,msg1")) == \
                         [kcm, msg1]
        assert f.mock_calls == \
                         [mock.call.add_and_parse(b"kcm,msg1")]
        assert n.mock_calls == [mock.call.decrypt("f_kcm"),
                                        mock.call.decrypt("f_msg1")]
        assert pr.mock_calls == [mock.call(kcm), mock.call(msg1)]
    n.mock_calls[:] = []
    f.mock_calls[:] = []

def test_large_frame():
    """
    Noise only allows 64KiB message, but the API allows up to 4GiB
    frames
    """
    if not NoiseConnection:
        import unittest
        raise unittest.SkipTest("noiseprotocol unavailable")
    # XXX could really benefit from some Hypothesis style
    # exploration of more cases .. but we don't already depend on
    # that library, so a future improvement

    @implementer(ITransport)
    class FakeTransport:
        """
        Record which write()s happen
        """
        def __init__(self):
            self.data = []  # list of bytes

        def write(self, data):
            self.data.append(data)

    # we build both sides of a connection so that underlying Noise
    # structures can be set up and paired properly. Essentially
    # this test is acting like the L2 Protocol object, and can
    # feed bytes to / from either side
    pake_secret = b"\x00" * 32
    transport0 = FakeTransport()
    transport1 = FakeTransport()
    noise0 = build_noise()
    noise0.set_psks(pake_secret)
    noise0.set_as_initiator()  # leader
    noise1 = build_noise()
    noise1.set_psks(pake_secret)
    noise1.set_as_responder()  # follower
    framer0 = _Framer(transport0, b"out prolog", b"in prolog")
    record0 = _Record(framer0, noise0, LEADER)
    record0.set_role_leader()
    record0.connectionMade()
    assert list(record0.add_and_unframe(b"in prolog")) == \
        []

    # note that in connector the prologues flip around depending
    # on who is the leader or follower
    framer1 = _Framer(transport1, b"in prolog", b"out prolog")
    record1 = _Record(framer1, noise1, FOLLOWER)
    record1.set_role_follower()
    record1.connectionMade()

    # the leader has now sent the prolog and the opening handshake
    # message, so we consume them on the follower side

    assert list(record1.add_and_unframe(transport0.data[0])) == [] # b"out prolog"
    assert list(record1.add_and_unframe(transport0.data[1])) == [Handshake()]

    # the follower sends the first KCM; the leader is waiting for
    # this and will complete the handshake after that
    record1.send_record(KCM())
    assert list(record0.add_and_unframe(transport1.data[1])) == \
        [Handshake()]
    record0.send_record(KCM())
    assert list(record1.add_and_unframe(transport0.data[2])) == \
        [KCM()]
    # both sides are now done their handshakes

    # Now, we send a message that's definitely bigger than a
    # single Noise message can deal with
    input_plaintext = b"\xff" * 65537
    record0.send_record(
        Data(
            seqnum=123,
            scid=456,
            data=input_plaintext,
        )
    )

    # ...despite its size the message should still be sent out as
    # a single Data, because _our_ framing handles 4-byte lengths
    assert len(transport0.data[3]) == \
        4 + 65537 + (16 * 2) + 1 + 4 + 4
    #   ^-- frame-length
    #       ^-- plaintext size
    #               ^-- Noise associated data, x2 noise packets
    #                          ^-- "data" kind, 0x04
    #                              ^-- subchannel-id
    #                                  ^-- sequence number
    #
    # (maybe we don't need the above assert, since if any of that
    # isn't true the next step will fail)

    # round-trip the data to the other side and ensure we get the
    # plaintext back
    outputs = list(
        record1.add_and_unframe(transport0.data[3])
    )
    assert len(outputs) == 1
    assert outputs[0] == \
        Data(
            seqnum=123,
            scid=456,
            data=input_plaintext,
        )



from unittest import mock
from zope.interface import alsoProvides
from ..._interfaces import IDilationManager, IWormhole


def mock_manager():
    m = mock.Mock()
    alsoProvides(m, IDilationManager)
    return m


def mock_wormhole():
    m = mock.Mock()
    alsoProvides(m, IWormhole)
    return m


def clear_mock_calls(*args):
    for a in args:
        a.mock_calls[:] = []



from twisted.internet.defer import Deferred
from twisted.internet.task import Clock

from unittest import mock

from pytest_twisted import ensureDeferred

from ..eventual import EventualQueue


class IntentionalError(Exception):
    pass


def test_eventually(observe_errors):
    c = Clock()
    eq = EventualQueue(c)
    c1 = mock.Mock()
    eq.eventually(c1, "arg1", "arg2", kwarg1="kw1")
    eq.eventually(c1, "arg3", "arg4", kwarg5="kw5")
    d2 = eq.fire_eventually()
    d3 = eq.fire_eventually("value")
    assert c1.mock_calls == []
    assert not d2.called
    assert not d3.called

    eq.flush_sync()
    assert c1.mock_calls == [
        mock.call("arg1", "arg2", kwarg1="kw1"),
        mock.call("arg3", "arg4", kwarg5="kw5")
    ]
    assert d2.result is None
    assert d3.result == "value"


def test_error(observe_errors):
    c = Clock()
    eq = EventualQueue(c)
    c1 = mock.Mock(side_effect=IntentionalError)
    eq.eventually(c1, "arg1", "arg2", kwarg1="kw1")
    assert c1.mock_calls == []

    eq.flush_sync()
    assert c1.mock_calls == [mock.call("arg1", "arg2", kwarg1="kw1")]
    observe_errors.flush(IntentionalError)


@ensureDeferred
async def test_flush(reactor, observe_errors):
    eq = EventualQueue(reactor)
    d1 = eq.fire_eventually()
    d2 = Deferred()

    def _more(res):
        eq.eventually(d2.callback, None)

    d1.addCallback(_more)
    await eq.flush()
    # d1 will fire, which will queue d2 to fire, and the flush() ought to
    # wait for d2 too
    assert d2.called



import json

from nacl.secret import SecretBox
from spake2 import SPAKE2_Symmetric
from zope.interface import directlyProvides, implementer

from unittest import mock

from .. import (__version__, _allocator, _boss, _code, _input, _key, _lister,
                _mailbox, _nameplate, _order, _receive, _rendezvous, _send,
                _terminator, errors, timing)
from .._interfaces import (IAllocator, IBoss, ICode, IDilator, IInput, IKey,
                           ILister, IMailbox, INameplate, IOrder, IReceive,
                           IRendezvousConnector, ISend, ITerminator, IWordlist,
                           ITorManager)
from .._key import derive_key, derive_phase_key, encrypt_data
from ..journal import ImmediateJournal
from .._status import WormholeStatus
from ..util import (bytes_to_dict, bytes_to_hexstr, dict_to_bytes,
                    hexstr_to_bytes, to_bytes)
import pytest
import pytest_twisted


@implementer(IWordlist)
class FakeWordList:
    def choose_words(self, length):
        return "-".join(["word"] * length)

    def get_completions(self, prefix):
        self._get_completions_prefix = prefix
        return self._completions


class Dummy:
    def __init__(self, name, events, iface, *meths, **kw):
        self.name = name
        self.events = events
        if iface:
            directlyProvides(self, iface)
        for meth in meths:
            self.mock(meth)
        self.retval = None
        for k, v in kw.items():
            setattr(self, k, v)

    def mock(self, meth):
        def log(*args):
            self.events.append((f"{self.name}.{meth}", ) + args)
            return self.retval

        setattr(self, meth, log)


def build_send():
    events = []
    s = _send.Send("side", timing.DebugTiming())
    m = Dummy("m", events, IMailbox, "add_message")
    s.wire(m)
    return s, m, events


def test_send_first():
    s, m, events = build_send()
    s.send("phase1", b"msg")
    assert events == []
    key = b"\x00" * 32
    nonce1 = b"\x00" * SecretBox.NONCE_SIZE
    with mock.patch("nacl.utils.random", side_effect=[nonce1]) as r:
        s.got_verified_key(key)
    assert r.mock_calls == [mock.call(SecretBox.NONCE_SIZE)]
    # print(bytes_to_hexstr(events[0][2]))
    enc1 = hexstr_to_bytes(
        "000000000000000000000000000000000000000000000000"
         "22f1a46c3c3496423c394621a2a5a8cf275b08")
    assert events == [("m.add_message", "phase1", enc1)]
    events[:] = []

    nonce2 = b"\x02" * SecretBox.NONCE_SIZE
    with mock.patch("nacl.utils.random", side_effect=[nonce2]) as r:
        s.send("phase2", b"msg")
    assert r.mock_calls == [mock.call(SecretBox.NONCE_SIZE)]
    enc2 = hexstr_to_bytes(
        "0202020202020202020202020202020202020202"
         "020202026660337c3eac6513c0dac9818b62ef16d9cd7e")
    assert events == [("m.add_message", "phase2", enc2)]

def test_key_first():
    s, m, events = build_send()
    key = b"\x00" * 32
    s.got_verified_key(key)
    assert events == []

    nonce1 = b"\x00" * SecretBox.NONCE_SIZE
    with mock.patch("nacl.utils.random", side_effect=[nonce1]) as r:
        s.send("phase1", b"msg")
    assert r.mock_calls == [mock.call(SecretBox.NONCE_SIZE)]
    enc1 = hexstr_to_bytes("00000000000000000000000000000000000000000000"
                            "000022f1a46c3c3496423c394621a2a5a8cf275b08")
    assert events == [("m.add_message", "phase1", enc1)]
    events[:] = []

    nonce2 = b"\x02" * SecretBox.NONCE_SIZE
    with mock.patch("nacl.utils.random", side_effect=[nonce2]) as r:
        s.send("phase2", b"msg")
    assert r.mock_calls == [mock.call(SecretBox.NONCE_SIZE)]
    enc2 = hexstr_to_bytes(
        "0202020202020202020202020202020202020"
         "202020202026660337c3eac6513c0dac9818b62ef16d9cd7e")
    assert events == [("m.add_message", "phase2", enc2)]


def build_order():
    events = []
    o = _order.Order("side", timing.DebugTiming())
    k = Dummy("k", events, IKey, "got_pake")
    r = Dummy("r", events, IReceive, "got_message")
    o.wire(k, r)
    return o, k, r, events


def test_in_order():
    o, k, r, events = build_order()
    o.got_message("side", "pake", b"body")
    assert events == [("k.got_pake", b"body")]  # right away
    o.got_message("side", "version", b"body")
    o.got_message("side", "1", b"body")
    assert events == [
        ("k.got_pake", b"body"),
        ("r.got_message", "side", "version", b"body"),
        ("r.got_message", "side", "1", b"body"),
    ]

def test_out_of_order():
    o, k, r, events = build_order()
    o.got_message("side", "version", b"body")
    assert events == []  # nothing yet
    o.got_message("side", "1", b"body")
    assert events == []  # nothing yet
    o.got_message("side", "pake", b"body")
    # got_pake is delivered first
    assert events == [
        ("k.got_pake", b"body"),
        ("r.got_message", "side", "version", b"body"),
        ("r.got_message", "side", "1", b"body"),
    ]


def build_receive():
    events = []
    r = _receive.Receive("side", timing.DebugTiming())
    b = Dummy("b", events, IBoss, "happy", "scared", "got_verifier",
              "got_message")
    s = Dummy("s", events, ISend, "got_verified_key")
    r.wire(b, s)
    return r, b, s, events


def test_good_receive():
    r, b, s, events = build_receive()
    key = b"key"
    r.got_key(key)
    assert events == []
    verifier = derive_key(key, b"wormhole:verifier")
    phase1_key = derive_phase_key(key, "side", "phase1")
    data1 = b"data1"
    good_body = encrypt_data(phase1_key, data1)
    r.got_message("side", "phase1", good_body)
    assert events == [
        ("s.got_verified_key", key),
        ("b.happy", ),
        ("b.got_verifier", verifier),
        ("b.got_message", "phase1", data1),
    ]

    phase2_key = derive_phase_key(key, "side", "phase2")
    data2 = b"data2"
    good_body = encrypt_data(phase2_key, data2)
    r.got_message("side", "phase2", good_body)
    assert events == [
        ("s.got_verified_key", key),
        ("b.happy", ),
        ("b.got_verifier", verifier),
        ("b.got_message", "phase1", data1),
        ("b.got_message", "phase2", data2),
    ]

def test_early_bad():
    r, b, s, events = build_receive()
    key = b"key"
    r.got_key(key)
    assert events == []
    phase1_key = derive_phase_key(key, "side", "bad")
    data1 = b"data1"
    bad_body = encrypt_data(phase1_key, data1)
    r.got_message("side", "phase1", bad_body)
    assert events == [
        ("b.scared", ),
    ]

    phase2_key = derive_phase_key(key, "side", "phase2")
    data2 = b"data2"
    good_body = encrypt_data(phase2_key, data2)
    r.got_message("side", "phase2", good_body)
    assert events == [
        ("b.scared", ),
    ]

def test_late_bad():
    r, b, s, events = build_receive()
    key = b"key"
    r.got_key(key)
    assert events == []
    verifier = derive_key(key, b"wormhole:verifier")
    phase1_key = derive_phase_key(key, "side", "phase1")
    data1 = b"data1"
    good_body = encrypt_data(phase1_key, data1)
    r.got_message("side", "phase1", good_body)
    assert events == [
        ("s.got_verified_key", key),
        ("b.happy", ),
        ("b.got_verifier", verifier),
        ("b.got_message", "phase1", data1),
    ]

    phase2_key = derive_phase_key(key, "side", "bad")
    data2 = b"data2"
    bad_body = encrypt_data(phase2_key, data2)
    r.got_message("side", "phase2", bad_body)
    assert events == [
        ("s.got_verified_key", key),
        ("b.happy", ),
        ("b.got_verifier", verifier),
        ("b.got_message", "phase1", data1),
        ("b.scared", ),
    ]
    r.got_message("side", "phase1", good_body)
    r.got_message("side", "phase2", bad_body)
    assert events == [
        ("s.got_verified_key", key),
        ("b.happy", ),
        ("b.got_verifier", verifier),
        ("b.got_message", "phase1", data1),
        ("b.scared", ),
    ]


def build_key():
    events = []
    k = _key.Key("appid", {}, "side", timing.DebugTiming())
    b = Dummy("b", events, IBoss, "scared", "got_key")
    m = Dummy("m", events, IMailbox, "add_message")
    r = Dummy("r", events, IReceive, "got_key")
    k.wire(b, m, r)
    return k, b, m, r, events


def test_good_key():
    k, b, m, r, events = build_key()
    code = "1-foo"
    k.got_code(code)
    assert len(events) == 1
    assert events[0][:2] == ("m.add_message", "pake")
    msg1_json = events[0][2].decode("utf-8")
    events[:] = []
    msg1 = json.loads(msg1_json)
    msg1_bytes = hexstr_to_bytes(msg1["pake_v1"])
    sp = SPAKE2_Symmetric(to_bytes(code), idSymmetric=to_bytes("appid"))
    msg2_bytes = sp.start()
    key2 = sp.finish(msg1_bytes)
    msg2 = dict_to_bytes({"pake_v1": bytes_to_hexstr(msg2_bytes)})
    k.got_pake(msg2)
    assert len(events) == 3, events
    assert events[0] == ("b.got_key", key2)
    assert events[1][:2] == ("m.add_message", "version")
    assert events[2] == ("r.got_key", key2)

def test_bad():
    k, b, m, r, events = build_key()
    code = "1-foo"
    k.got_code(code)
    assert len(events) == 1
    assert events[0][:2] == ("m.add_message", "pake")
    pake_1_json = events[0][2].decode("utf-8")
    pake_1 = json.loads(pake_1_json)
    assert list(pake_1.keys()) == \
                     ["pake_v1"]  # value is PAKE stuff
    events[:] = []
    bad_pake_d = {"not_pake_v1": "stuff"}
    k.got_pake(dict_to_bytes(bad_pake_d))
    assert events == [("b.scared", )]

def test_reversed():
    # A receiver using input_code() will choose the nameplate first, then
    # the rest of the code. Once the nameplate is selected, we'll claim
    # it and open the mailbox, which will cause the senders PAKE to
    # arrive before the code has been set. Key() is supposed to stash the
    # PAKE message until the code is set (allowing the PAKE computation
    # to finish). This test exercises that PAKE-then-code sequence.
    k, b, m, r, events = build_key()
    code = "1-foo"

    sp = SPAKE2_Symmetric(to_bytes(code), idSymmetric=to_bytes("appid"))
    msg2_bytes = sp.start()
    msg2 = dict_to_bytes({"pake_v1": bytes_to_hexstr(msg2_bytes)})
    k.got_pake(msg2)
    assert len(events) == 0

    k.got_code(code)
    assert len(events) == 4
    assert events[0][:2] == ("m.add_message", "pake")
    msg1_json = events[0][2].decode("utf-8")
    msg1 = json.loads(msg1_json)
    msg1_bytes = hexstr_to_bytes(msg1["pake_v1"])
    key2 = sp.finish(msg1_bytes)
    assert events[1] == ("b.got_key", key2)
    assert events[2][:2] == ("m.add_message", "version")
    assert events[3] == ("r.got_key", key2)


def build_code():
    events = []
    c = _code.Code(timing.DebugTiming())
    b = Dummy("b", events, IBoss, "got_code")
    a = Dummy("a", events, IAllocator, "allocate")
    n = Dummy("n", events, INameplate, "set_nameplate")
    k = Dummy("k", events, IKey, "got_code")
    i = Dummy("i", events, IInput, "start")
    c.wire(b, a, n, k, i)
    return c, b, a, n, k, i, events


def test_set_code():
    c, b, a, n, k, i, events = build_code()
    c.set_code("1-code")
    assert events == [
        ("n.set_nameplate", "1"),
        ("b.got_code", "1-code"),
        ("k.got_code", "1-code"),
    ]

def test_set_code_invalid():
    c, b, a, n, k, i, events = build_code()
    with pytest.raises(errors.KeyFormatError) as e:
        c.set_code("1-code ")
    assert str(e.value) == "Code '1-code ' contains spaces."
    with pytest.raises(errors.KeyFormatError) as e:
        c.set_code(" 1-code")
    assert str(e.value) == "Code ' 1-code' contains spaces."
    with pytest.raises(errors.KeyFormatError) as e:
        c.set_code("code-code")
    assert str(e.value) == \
        "Nameplate 'code' must be numeric, with no spaces."

    # it should still be possible to use the wormhole at this point
    c.set_code("1-code")
    assert events == [
        ("n.set_nameplate", "1"),
        ("b.got_code", "1-code"),
        ("k.got_code", "1-code"),
    ]

def test_allocate_code():
    c, b, a, n, k, i, events = build_code()
    wl = FakeWordList()
    c.allocate_code(2, wl)
    assert events == [("a.allocate", 2, wl)]
    events[:] = []
    c.allocated("1", "1-code")
    assert events == [
        ("n.set_nameplate", "1"),
        ("b.got_code", "1-code"),
        ("k.got_code", "1-code"),
    ]

def test_input_code():
    c, b, a, n, k, i, events = build_code()
    c.input_code()
    assert events == [("i.start", )]
    events[:] = []
    c.got_nameplate("1")
    assert events == [
        ("n.set_nameplate", "1"),
    ]
    events[:] = []
    c.finished_input("1-code")
    assert events == [
        ("b.got_code", "1-code"),
        ("k.got_code", "1-code"),
    ]


def build_input():
    events = []
    i = _input.Input(timing.DebugTiming())
    code = Dummy("c", events, ICode, "got_nameplate", "finished_input")
    lister = Dummy("l", events, ILister, "refresh")
    i.wire(code, lister)
    return i, code, lister, events


def test_ignore_completion():
    i, c, lister, events = build_input()
    helper = i.start()
    assert isinstance(helper, _input.Helper)
    assert events == [("l.refresh", )]
    events[:] = []
    with pytest.raises(errors.MustChooseNameplateFirstError):
        helper.choose_words("word-word")
    helper.choose_nameplate("1")
    assert events == [("c.got_nameplate", "1")]
    events[:] = []
    with pytest.raises(errors.AlreadyChoseNameplateError):
        helper.choose_nameplate("2")
    helper.choose_words("word-word")
    with pytest.raises(errors.AlreadyChoseWordsError):
        helper.choose_words("word-word")
    assert events == [("c.finished_input", "1-word-word")]

def test_bad_nameplate():
    i, c, lister, events = build_input()
    helper = i.start()
    assert isinstance(helper, _input.Helper)
    assert events == [("l.refresh", )]
    events[:] = []
    with pytest.raises(errors.MustChooseNameplateFirstError):
        helper.choose_words("word-word")
    with pytest.raises(errors.KeyFormatError):
        helper.choose_nameplate(" 1")
    # should still work afterwards
    helper.choose_nameplate("1")
    assert events == [("c.got_nameplate", "1")]
    events[:] = []
    with pytest.raises(errors.AlreadyChoseNameplateError):
        helper.choose_nameplate("2")
    helper.choose_words("word-word")
    with pytest.raises(errors.AlreadyChoseWordsError):
        helper.choose_words("word-word")
    assert events == [("c.finished_input", "1-word-word")]


@pytest_twisted.ensureDeferred
async def test_with_completion():
    i, c, lister, events = build_input()
    helper = i.start()
    assert isinstance(helper, _input.Helper)
    assert events == [("l.refresh", )]
    events[:] = []
    d = helper.when_wordlist_is_available()
    assert not d.called
    helper.refresh_nameplates()
    assert events == [("l.refresh", )]
    events[:] = []
    with pytest.raises(errors.MustChooseNameplateFirstError):
        helper.get_word_completions("prefix")
    i.got_nameplates({"1", "12", "34", "35", "367"})
    assert not d.called
    assert helper.get_nameplate_completions("") == \
        {"1-", "12-", "34-", "35-", "367-"}
    assert helper.get_nameplate_completions("1") == {"1-", "12-"}
    assert helper.get_nameplate_completions("2") == set()
    assert helper.get_nameplate_completions("3") == {"34-", "35-", "367-"}
    helper.choose_nameplate("34")
    with pytest.raises(errors.AlreadyChoseNameplateError):
        helper.refresh_nameplates()
    with pytest.raises(errors.AlreadyChoseNameplateError):
        helper.get_nameplate_completions("1")
    assert events == [("c.got_nameplate", "34")]
    events[:] = []
    # no wordlist yet
    assert not d.called
    assert helper.get_word_completions("") == set()
    wl = FakeWordList()
    i.got_wordlist(wl)
    assert await d is None
    # a new Deferred should fire right away
    d = helper.when_wordlist_is_available()
    assert await d is None

    wl._completions = {"abc-", "abcd-", "ae-"}
    assert helper.get_word_completions("a") == wl._completions
    assert wl._get_completions_prefix == "a"
    with pytest.raises(errors.AlreadyChoseNameplateError):
        helper.refresh_nameplates()
    with pytest.raises(errors.AlreadyChoseNameplateError):
        helper.get_nameplate_completions("1")
    helper.choose_words("word-word")
    with pytest.raises(errors.AlreadyChoseWordsError):
        helper.get_word_completions("prefix")
    with pytest.raises(errors.AlreadyChoseWordsError):
        helper.choose_words("word-word")
    assert events == [("c.finished_input", "34-word-word")]


def build_lister():
    events = []
    lister = _lister.Lister(timing.DebugTiming())
    rc = Dummy("rc", events, IRendezvousConnector, "tx_list")
    i = Dummy("i", events, IInput, "got_nameplates")
    lister.wire(rc, i)
    return lister, rc, i, events


def test_connect_first_lister():
    lister, rc, i, events = build_lister()
    lister.connected()
    lister.lost()
    lister.connected()
    assert events == []
    lister.refresh()
    assert events == [
        ("rc.tx_list", ),
    ]
    events[:] = []
    lister.rx_nameplates({"1", "2", "3"})
    assert events == [
        ("i.got_nameplates", {"1", "2", "3"}),
    ]
    events[:] = []
    # now we're satisfied: disconnecting and reconnecting won't ask again
    lister.lost()
    lister.connected()
    assert events == []

    # but if we're told to refresh, we'll do so
    lister.refresh()
    assert events == [
        ("rc.tx_list", ),
    ]

def test_connect_first_ask_twice():
    lister, rc, i, events = build_lister()
    lister.connected()
    assert events == []
    lister.refresh()
    lister.refresh()
    assert events == [
        ("rc.tx_list", ),
        ("rc.tx_list", ),
    ]
    lister.rx_nameplates({"1", "2", "3"})
    assert events == [
        ("rc.tx_list", ),
        ("rc.tx_list", ),
        ("i.got_nameplates", {"1", "2", "3"}),
    ]
    lister.rx_nameplates({"1", "2", "3", "4"})
    assert events == [
        ("rc.tx_list", ),
        ("rc.tx_list", ),
        ("i.got_nameplates", {"1", "2", "3"}),
        ("i.got_nameplates", {"1", "2", "3", "4"}),
    ]

def test_reconnect():
    lister, rc, i, events = build_lister()
    lister.refresh()
    lister.connected()
    assert events == [
        ("rc.tx_list", ),
    ]
    events[:] = []
    lister.lost()
    lister.connected()
    assert events == [
        ("rc.tx_list", ),
    ]

def test_refresh_first():
    lister, rc, i, events = build_lister()
    lister.refresh()
    assert events == []
    lister.connected()
    assert events == [
        ("rc.tx_list", ),
    ]
    lister.rx_nameplates({"1", "2", "3"})
    assert events == [
        ("rc.tx_list", ),
        ("i.got_nameplates", {"1", "2", "3"}),
    ]

def test_unrefreshed():
    lister, rc, i, events = build_lister()
    assert events == []
    # we receive a spontaneous rx_nameplates, without asking
    lister.connected()
    assert events == []
    lister.rx_nameplates({"1", "2", "3"})
    assert events == [
        ("i.got_nameplates", {"1", "2", "3"}),
    ]


def build_allocator():
    events = []
    a = _allocator.Allocator(timing.DebugTiming())
    rc = Dummy("rc", events, IRendezvousConnector, "tx_allocate")
    c = Dummy("c", events, ICode, "allocated")
    a.wire(rc, c)
    return a, rc, c, events


def test_no_allocation():
    a, rc, c, events = build_allocator()
    a.connected()
    assert events == []

def test_allocate_first():
    a, rc, c, events = build_allocator()
    a.allocate(2, FakeWordList())
    assert events == []
    a.connected()
    assert events == [("rc.tx_allocate", )]
    events[:] = []
    a.lost()
    a.connected()
    assert events == [
        ("rc.tx_allocate", ),
    ]
    events[:] = []
    a.rx_allocated("1")
    assert events == [
        ("c.allocated", "1", "1-word-word"),
    ]

def test_connect_first_allocator():
    a, rc, c, events = build_allocator()
    a.connected()
    assert events == []
    a.allocate(2, FakeWordList())
    assert events == [("rc.tx_allocate", )]
    events[:] = []
    a.lost()
    a.connected()
    assert events == [
        ("rc.tx_allocate", ),
    ]
    events[:] = []
    a.rx_allocated("1")
    assert events == [
        ("c.allocated", "1", "1-word-word"),
    ]


def build_nameplate():
    events = []
    n = _nameplate.Nameplate(lambda **kw: None)
    m = Dummy("m", events, IMailbox, "got_mailbox")
    i = Dummy("i", events, IInput, "got_wordlist")
    rc = Dummy("rc", events, IRendezvousConnector, "tx_claim",
               "tx_release")
    t = Dummy("t", events, ITerminator, "nameplate_done")
    n.wire(m, i, rc, t)
    return n, m, i, rc, t, events


def test_set_invalid():
    n, m, i, rc, t, events = build_nameplate()
    with pytest.raises(errors.KeyFormatError) as e:
        n.set_nameplate(" 1")
    assert str(e.value) == \
        "Nameplate ' 1' must be numeric, with no spaces."
    with pytest.raises(errors.KeyFormatError) as e:
        n.set_nameplate("one")
    assert str(e.value) == \
        "Nameplate 'one' must be numeric, with no spaces."

    # wormhole should still be usable
    n.set_nameplate("1")
    assert events == []
    n.connected()
    assert events == [("rc.tx_claim", "1")]

def test_set_first():
    # connection remains up throughout
    n, m, i, rc, t, events = build_nameplate()
    n.set_nameplate("1")
    assert events == []
    n.connected()
    assert events == [("rc.tx_claim", "1")]
    events[:] = []

    wl = object()
    with mock.patch("wormhole._nameplate.PGPWordList", return_value=wl):
        n.rx_claimed("mbox1")
    assert events == [
        ("i.got_wordlist", wl),
        ("m.got_mailbox", "mbox1"),
    ]
    events[:] = []

    n.release()
    assert events == [("rc.tx_release", "1")]
    events[:] = []

    n.rx_released()
    assert events == [("t.nameplate_done", )]

def test_connect_first_nameplate():
    # connection remains up throughout
    n, m, i, rc, t, events = build_nameplate()
    n.connected()
    assert events == []

    n.set_nameplate("1")
    assert events == [("rc.tx_claim", "1")]
    events[:] = []

    wl = object()
    with mock.patch("wormhole._nameplate.PGPWordList", return_value=wl):
        n.rx_claimed("mbox1")
    assert events == [
        ("i.got_wordlist", wl),
        ("m.got_mailbox", "mbox1"),
    ]
    events[:] = []

    n.release()
    assert events == [("rc.tx_release", "1")]
    events[:] = []

    n.rx_released()
    assert events == [("t.nameplate_done", )]

def test_reconnect_while_claiming():
    # connection bounced while waiting for rx_claimed
    n, m, i, rc, t, events = build_nameplate()
    n.connected()
    assert events == []

    n.set_nameplate("1")
    assert events == [("rc.tx_claim", "1")]
    events[:] = []

    n.lost()
    n.connected()
    assert events == [("rc.tx_claim", "1")]

def test_reconnect_while_claimed():
    # connection bounced while claimed: no retransmits should be sent
    n, m, i, rc, t, events = build_nameplate()
    n.connected()
    assert events == []

    n.set_nameplate("1")
    assert events == [("rc.tx_claim", "1")]
    events[:] = []

    wl = object()
    with mock.patch("wormhole._nameplate.PGPWordList", return_value=wl):
        n.rx_claimed("mbox1")
    assert events == [
        ("i.got_wordlist", wl),
        ("m.got_mailbox", "mbox1"),
    ]
    events[:] = []

    n.lost()
    n.connected()
    assert events == []

def test_reconnect_while_releasing():
    # connection bounced while waiting for rx_released
    n, m, i, rc, t, events = build_nameplate()
    n.connected()
    assert events == []

    n.set_nameplate("1")
    assert events == [("rc.tx_claim", "1")]
    events[:] = []

    wl = object()
    with mock.patch("wormhole._nameplate.PGPWordList", return_value=wl):
        n.rx_claimed("mbox1")
    assert events == [
        ("i.got_wordlist", wl),
        ("m.got_mailbox", "mbox1"),
    ]
    events[:] = []

    n.release()
    assert events == [("rc.tx_release", "1")]
    events[:] = []

    n.lost()
    n.connected()
    assert events == [("rc.tx_release", "1")]

def test_reconnect_while_done():
    # connection bounces after we're done
    n, m, i, rc, t, events = build_nameplate()
    n.connected()
    assert events == []

    n.set_nameplate("1")
    assert events == [("rc.tx_claim", "1")]
    events[:] = []

    wl = object()
    with mock.patch("wormhole._nameplate.PGPWordList", return_value=wl):
        n.rx_claimed("mbox1")
    assert events == [
        ("i.got_wordlist", wl),
        ("m.got_mailbox", "mbox1"),
    ]
    events[:] = []

    n.release()
    assert events == [("rc.tx_release", "1")]
    events[:] = []

    n.rx_released()
    assert events == [("t.nameplate_done", )]
    events[:] = []

    n.lost()
    n.connected()
    assert events == []


def test_close_while_idle_nameplate():
    n, m, i, rc, t, events = build_nameplate()
    n.close()
    assert events == [("t.nameplate_done", )]


def test_close_while_idle_connected():
    n, m, i, rc, t, events = build_nameplate()
    n.connected()
    assert events == []
    n.close()
    assert events == [("t.nameplate_done", )]


def test_close_while_unclaimed():
    n, m, i, rc, t, events = build_nameplate()
    n.set_nameplate("1")
    n.close()  # before ever being connected
    assert events == [("t.nameplate_done", )]

def test_close_while_claiming():
    n, m, i, rc, t, events = build_nameplate()
    n.set_nameplate("1")
    assert events == []
    n.connected()
    assert events == [("rc.tx_claim", "1")]
    events[:] = []

    n.close()
    assert events == [("rc.tx_release", "1")]
    events[:] = []

    n.rx_released()
    assert events == [("t.nameplate_done", )]

def test_close_while_claiming_but_disconnected():
    n, m, i, rc, t, events = build_nameplate()
    n.set_nameplate("1")
    assert events == []
    n.connected()
    assert events == [("rc.tx_claim", "1")]
    events[:] = []

    n.lost()
    n.close()
    assert events == []
    # we're now waiting for a connection, so we can release the nameplate
    n.connected()
    assert events == [("rc.tx_release", "1")]
    events[:] = []

    n.rx_released()
    assert events == [("t.nameplate_done", )]

def test_close_while_claimed():
    n, m, i, rc, t, events = build_nameplate()
    n.set_nameplate("1")
    assert events == []
    n.connected()
    assert events == [("rc.tx_claim", "1")]
    events[:] = []

    wl = object()
    with mock.patch("wormhole._nameplate.PGPWordList", return_value=wl):
        n.rx_claimed("mbox1")
    assert events == [
        ("i.got_wordlist", wl),
        ("m.got_mailbox", "mbox1"),
    ]
    events[:] = []

    n.close()
    # this path behaves just like a deliberate release()
    assert events == [("rc.tx_release", "1")]
    events[:] = []

    n.rx_released()
    assert events == [("t.nameplate_done", )]

def test_close_while_claimed_but_disconnected():
    n, m, i, rc, t, events = build_nameplate()
    n.set_nameplate("1")
    assert events == []
    n.connected()
    assert events == [("rc.tx_claim", "1")]
    events[:] = []

    wl = object()
    with mock.patch("wormhole._nameplate.PGPWordList", return_value=wl):
        n.rx_claimed("mbox1")
    assert events == [
        ("i.got_wordlist", wl),
        ("m.got_mailbox", "mbox1"),
    ]
    events[:] = []

    n.lost()
    n.close()
    # we're now waiting for a connection, so we can release the nameplate
    n.connected()
    assert events == [("rc.tx_release", "1")]
    events[:] = []

    n.rx_released()
    assert events == [("t.nameplate_done", )]

def test_close_while_releasing():
    n, m, i, rc, t, events = build_nameplate()
    n.set_nameplate("1")
    assert events == []
    n.connected()
    assert events == [("rc.tx_claim", "1")]
    events[:] = []

    wl = object()
    with mock.patch("wormhole._nameplate.PGPWordList", return_value=wl):
        n.rx_claimed("mbox1")
    assert events == [
        ("i.got_wordlist", wl),
        ("m.got_mailbox", "mbox1"),
    ]
    events[:] = []

    n.release()
    assert events == [("rc.tx_release", "1")]
    events[:] = []

    n.close()  # ignored, we're already on our way out the door
    assert events == []
    n.rx_released()
    assert events == [("t.nameplate_done", )]

def test_close_while_releasing_but_disconnecteda():
    n, m, i, rc, t, events = build_nameplate()
    n.set_nameplate("1")
    assert events == []
    n.connected()
    assert events == [("rc.tx_claim", "1")]
    events[:] = []

    wl = object()
    with mock.patch("wormhole._nameplate.PGPWordList", return_value=wl):
        n.rx_claimed("mbox1")
    assert events == [
        ("i.got_wordlist", wl),
        ("m.got_mailbox", "mbox1"),
    ]
    events[:] = []

    n.release()
    assert events == [("rc.tx_release", "1")]
    events[:] = []

    n.lost()
    n.close()
    # we must retransmit the tx_release when we reconnect
    assert events == []

    n.connected()
    assert events == [("rc.tx_release", "1")]
    events[:] = []

    n.rx_released()
    assert events == [("t.nameplate_done", )]

def test_close_while_done():
    # connection remains up throughout
    n, m, i, rc, t, events = build_nameplate()
    n.connected()
    assert events == []

    n.set_nameplate("1")
    assert events == [("rc.tx_claim", "1")]
    events[:] = []

    wl = object()
    with mock.patch("wormhole._nameplate.PGPWordList", return_value=wl):
        n.rx_claimed("mbox1")
    assert events == [
        ("i.got_wordlist", wl),
        ("m.got_mailbox", "mbox1"),
    ]
    events[:] = []

    n.release()
    assert events == [("rc.tx_release", "1")]
    events[:] = []

    n.rx_released()
    assert events == [("t.nameplate_done", )]
    events[:] = []

    n.close()  # NOP
    assert events == []

def test_close_while_done_but_disconnected():
    # connection remains up throughout
    n, m, i, rc, t, events = build_nameplate()
    n.connected()
    assert events == []

    n.set_nameplate("1")
    assert events == [("rc.tx_claim", "1")]
    events[:] = []

    wl = object()
    with mock.patch("wormhole._nameplate.PGPWordList", return_value=wl):
        n.rx_claimed("mbox1")
    assert events == [
        ("i.got_wordlist", wl),
        ("m.got_mailbox", "mbox1"),
    ]
    events[:] = []

    n.release()
    assert events == [("rc.tx_release", "1")]
    events[:] = []

    n.rx_released()
    assert events == [("t.nameplate_done", )]
    events[:] = []

    n.lost()
    n.close()  # NOP
    assert events == []


def build_mailbox():
    events = []
    m = _mailbox.Mailbox("side1")
    n = Dummy("n", events, INameplate, "release")
    rc = Dummy("rc", events, IRendezvousConnector, "tx_add", "tx_open",
               "tx_close")
    o = Dummy("o", events, IOrder, "got_message")
    t = Dummy("t", events, ITerminator, "mailbox_done")
    m.wire(n, rc, o, t)
    return m, n, rc, o, t, events

    # TODO: test moods

def assert_events(events, initial_events, tx_add_events):
    assert len(events) == \
        len(initial_events) + len(tx_add_events), events
    assert events[:len(initial_events)] == initial_events
    assert set(events[len(initial_events):]) == tx_add_events


def test_connect_first_mailbox():  # connect before got_mailbox
    m, n, rc, o, t, events = build_mailbox()
    m.add_message("phase1", b"msg1")
    assert events == []

    m.connected()
    assert events == []

    m.got_mailbox("mbox1")
    assert events == [("rc.tx_open", "mbox1"),
                              ("rc.tx_add", "phase1", b"msg1")]
    events[:] = []

    m.add_message("phase2", b"msg2")
    assert events == [("rc.tx_add", "phase2", b"msg2")]
    events[:] = []

    # bouncing the connection should retransmit everything, even the open()
    m.lost()
    assert events == []
    # and messages sent while here should be queued
    m.add_message("phase3", b"msg3")
    assert events == []

    m.connected()
    # the other messages are allowed to be sent in any order
    assert_events(
        events, [("rc.tx_open", "mbox1")], {
            ("rc.tx_add", "phase1", b"msg1"),
            ("rc.tx_add", "phase2", b"msg2"),
            ("rc.tx_add", "phase3", b"msg3"),
        })
    events[:] = []

    m.rx_message("side1", "phase1",
                 b"msg1")  # echo of our message, dequeue
    assert events == []

    m.lost()
    m.connected()
    assert_events(events, [("rc.tx_open", "mbox1")], {
        ("rc.tx_add", "phase2", b"msg2"),
        ("rc.tx_add", "phase3", b"msg3"),
    })
    events[:] = []

    # a new message from the peer gets delivered, and the Nameplate is
    # released since the message proves that our peer opened the Mailbox
    # and therefore no longer needs the Nameplate
    m.rx_message("side2", "phase1", b"msg1them")  # new message from peer
    assert events == [
        ("n.release", ),
        ("o.got_message", "side2", "phase1", b"msg1them"),
    ]
    events[:] = []

    # we de-duplicate peer messages, but still re-release the nameplate
    # since Nameplate is smart enough to ignore that
    m.rx_message("side2", "phase1", b"msg1them")
    assert events == [
        ("n.release", ),
    ]
    events[:] = []

    m.close("happy")
    assert events == [("rc.tx_close", "mbox1", "happy")]
    events[:] = []

    # while closing, we ignore a lot
    m.add_message("phase-late", b"late")
    m.rx_message("side1", "phase2", b"msg2")
    m.close("happy")
    assert events == []

    # bouncing the connection forces a retransmit of the tx_close
    m.lost()
    assert events == []
    m.connected()
    assert events == [("rc.tx_close", "mbox1", "happy")]
    events[:] = []

    m.rx_closed()
    assert events == [("t.mailbox_done", )]
    events[:] = []

    # while closed, we ignore everything
    m.add_message("phase-late", b"late")
    m.rx_message("side1", "phase2", b"msg2")
    m.close("happy")
    m.lost()
    m.connected()
    assert events == []

def test_mailbox_first():  # got_mailbox before connect
    m, n, rc, o, t, events = build_mailbox()
    m.add_message("phase1", b"msg1")
    assert events == []

    m.got_mailbox("mbox1")
    m.add_message("phase2", b"msg2")
    assert events == []

    m.connected()

    assert_events(events, [("rc.tx_open", "mbox1")], {
        ("rc.tx_add", "phase1", b"msg1"),
        ("rc.tx_add", "phase2", b"msg2"),
    })


def test_close_while_idle():
    m, n, rc, o, t, events = build_mailbox()
    m.close("happy")
    assert events == [("t.mailbox_done", )]


def test_close_while_idle_but_connected():
    m, n, rc, o, t, events = build_mailbox()
    m.connected()
    m.close("happy")
    assert events == [("t.mailbox_done", )]

def test_close_while_mailbox_disconnected():
    m, n, rc, o, t, events = build_mailbox()
    m.got_mailbox("mbox1")
    m.close("happy")
    assert events == [("t.mailbox_done", )]

def test_close_while_reconnecting():
    m, n, rc, o, t, events = build_mailbox()
    m.got_mailbox("mbox1")
    m.connected()
    assert events == [("rc.tx_open", "mbox1")]
    events[:] = []

    m.lost()
    assert events == []
    m.close("happy")
    assert events == []
    # we now wait to connect, so we can send the tx_close

    m.connected()
    assert events == [("rc.tx_close", "mbox1", "happy")]
    events[:] = []

    m.rx_closed()
    assert events == [("t.mailbox_done", )]
    events[:] = []


def build_terminator():
    events = []
    t = _terminator.Terminator()
    b = Dummy("b", events, IBoss, "closed")
    rc = Dummy("rc", events, IRendezvousConnector, "stop")
    n = Dummy("n", events, INameplate, "close")
    m = Dummy("m", events, IMailbox, "close")
    d = Dummy("d", events, IDilator, "stop")
    t.wire(b, rc, n, m, d)
    return t, b, rc, n, m, events


# there are three events, and we need to test all orderings of them
def _do_test(ev1, ev2, ev3):
    t, b, rc, n, m, events = build_terminator()
    input_events = {
        "mailbox": lambda: t.mailbox_done(),
        "nameplate": lambda: t.nameplate_done(),
        "rc": lambda: t.close("happy"),
    }
    close_events = [
        ("n.close", ),
        ("m.close", "happy"),
    ]

    if ev1 == "mailbox":
        close_events.remove(("m.close", "happy"))
    elif ev1 == "nameplate":
        close_events.remove(("n.close",))

    input_events[ev1]()
    expected = []
    if ev1 == "rc":
        expected.extend(close_events)
    assert events == expected
    events[:] = []

    if ev2 == "mailbox":
        close_events.remove(("m.close", "happy"))
    elif ev2 == "nameplate":
        close_events.remove(("n.close",))

    input_events[ev2]()
    expected = []
    if ev2 == "rc":
        expected.extend(close_events)
    assert events == expected
    events[:] = []

    if ev3 == "mailbox":
        close_events.remove(("m.close", "happy"))
    elif ev3 == "nameplate":
        close_events.remove(("n.close",))

    input_events[ev3]()
    expected = []
    if ev3 == "rc":
        expected.extend(close_events)
    expected.append(("rc.stop", ))
    assert events == expected
    events[:] = []

    t.stoppedRC()
    assert events == [("d.stop", )]
    events[:] = []

    t.stoppedD()
    assert events == [("b.closed", )]

def test_terminate():
    _do_test("mailbox", "nameplate", "rc")
    _do_test("mailbox", "rc", "nameplate")
    _do_test("nameplate", "mailbox", "rc")
    _do_test("nameplate", "rc", "mailbox")
    _do_test("rc", "nameplate", "mailbox")
    _do_test("rc", "mailbox", "nameplate")


# TODO: test moods


class MockBoss(_boss.Boss):
    def __attrs_post_init__(self):
        self._current_wormhole_status = WormholeStatus()
        # self._build_workers()
        self._init_other_state()


def build_boss():
    events = []
    wormhole = Dummy("w", events, None, "got_welcome", "got_code",
                     "got_key", "got_verifier", "got_versions", "received",
                     "closed")
    versions = {"app": "version1"}
    reactor = None
    eq = None
    cooperator = None
    journal = ImmediateJournal()
    tor_manager = None
    client_version = ("python", __version__)
    b = MockBoss(wormhole, "side", "url", "appid", versions,
                 client_version, reactor, eq, cooperator, journal,
                 tor_manager,
                 timing.DebugTiming())
    b._T = Dummy("t", events, ITerminator, "close")
    b._S = Dummy("s", events, ISend, "send")
    b._RC = Dummy("rc", events, IRendezvousConnector, "start")
    b._C = Dummy("c", events, ICode, "allocate_code", "input_code",
                 "set_code")
    b._D = Dummy("d", events, IDilator, "got_wormhole_versions", "got_key", _manager=None)
    return b, events

def test_boss_basic():
    b, events = build_boss()
    b.set_code("1-code")
    assert events == [("c.set_code", "1-code")]
    events[:] = []

    b.got_code("1-code")
    assert events == [("w.got_code", "1-code")]
    events[:] = []

    welcome = {"howdy": "how are ya"}
    b.rx_welcome(welcome)
    assert events == [
        ("w.got_welcome", welcome),
    ]
    events[:] = []

    # pretend a peer message was correctly decrypted
    b.got_key(b"key")
    b.happy()
    b.got_verifier(b"verifier")
    b.got_message("version", b"{}")
    b.got_message("0", b"msg1")
    assert events == [
        ("w.got_key", b"key"),
        ("d.got_key", b"key"),
        ("w.got_verifier", b"verifier"),
        ("d.got_wormhole_versions", {}),
        ("w.got_versions", {}),
        ("w.received", b"msg1"),
    ]
    events[:] = []

    b.send(b"msg2")
    assert events == [("s.send", "0", b"msg2")]
    events[:] = []

    b.close()
    assert events == [("t.close", "happy")]
    events[:] = []

    b.closed()
    assert events == [("w.closed", "happy")]

def test_unwelcome():
    b, events = build_boss()
    unwelcome = {"error": "go away"}
    b.rx_welcome(unwelcome)
    assert events == [("t.close", "unwelcome")]

def test_lonely():
    b, events = build_boss()
    b.set_code("1-code")
    assert events == [("c.set_code", "1-code")]
    events[:] = []

    b.got_code("1-code")
    assert events == [("w.got_code", "1-code")]
    events[:] = []

    b.close()
    assert events == [("t.close", "lonely")]
    events[:] = []

    b.closed()
    assert len(events) == 1, events
    assert events[0][0] == "w.closed"
    assert isinstance(events[0][1], errors.LonelyError)

def test_server_error():
    b, events = build_boss()
    b.set_code("1-code")
    assert events == [("c.set_code", "1-code")]
    events[:] = []

    orig = {}
    b.rx_error("server-error-msg", orig)
    assert events == [("t.close", "errory")]
    events[:] = []

    b.closed()
    assert len(events) == 1, events
    assert events[0][0] == "w.closed"
    assert isinstance(events[0][1], errors.ServerError)
    assert events[0][1].args[0] == "server-error-msg"

def test_internal_error():
    b, events = build_boss()
    b.set_code("1-code")
    assert events == [("c.set_code", "1-code")]
    events[:] = []

    b.error(ValueError("catch me"))
    assert len(events) == 1, events
    assert events[0][0] == "w.closed"
    assert isinstance(events[0][1], ValueError)
    assert events[0][1].args[0] == "catch me"

def test_close_early():
    b, events = build_boss()
    b.set_code("1-code")
    assert events == [("c.set_code", "1-code")]
    events[:] = []

    b.close()  # before even w.got_code
    assert events == [("t.close", "lonely")]
    events[:] = []

    b.closed()
    assert len(events) == 1, events
    assert events[0][0] == "w.closed"
    assert isinstance(events[0][1], errors.LonelyError)

def test_error_while_closing():
    b, events = build_boss()
    b.set_code("1-code")
    assert events == [("c.set_code", "1-code")]
    events[:] = []

    b.close()
    assert events == [("t.close", "lonely")]
    events[:] = []

    b.error(ValueError("oops"))
    assert len(events) == 1, events
    assert events[0][0] == "w.closed"
    assert isinstance(events[0][1], ValueError)

def test_scary_version():
    b, events = build_boss()
    b.set_code("1-code")
    assert events == [("c.set_code", "1-code")]
    events[:] = []

    b.got_code("1-code")
    assert events == [("w.got_code", "1-code")]
    events[:] = []

    b.scared()
    assert events == [("t.close", "scary")]
    events[:] = []

    b.closed()
    assert len(events) == 1, events
    assert events[0][0] == "w.closed"
    assert isinstance(events[0][1], errors.WrongPasswordError)

def test_scary_phase():
    b, events = build_boss()
    b.set_code("1-code")
    assert events == [("c.set_code", "1-code")]
    events[:] = []

    b.got_code("1-code")
    assert events == [("w.got_code", "1-code")]
    events[:] = []

    b.happy()  # phase=version

    b.scared()  # phase=0
    assert events == [("t.close", "scary")]
    events[:] = []

    b.closed()
    assert len(events) == 1, events
    assert events[0][0] == "w.closed"
    assert isinstance(events[0][1], errors.WrongPasswordError)


def test_unknown_phase(observe_errors):
    b, events = build_boss()
    b.set_code("1-code")
    assert events == [("c.set_code", "1-code")]
    events[:] = []

    b.got_code("1-code")
    assert events == [("w.got_code", "1-code")]
    events[:] = []

    b.happy()  # phase=version

    b.got_message("unknown-phase", b"spooky")
    assert events == []
    observe_errors.flush(errors._UnknownPhaseError)


def test_set_code_bad_format():
    b, events = build_boss()
    with pytest.raises(errors.KeyFormatError):
        b.set_code("1 code")
    # wormhole should still be usable
    b.set_code("1-code")
    assert events == [("c.set_code", "1-code")]


def test_set_code_twice():
    b, events = build_boss()
    b.set_code("1-code")
    with pytest.raises(errors.OnlyOneCodeError):
        b.set_code("1-code")


def test_input_code_boss():
    b, events = build_boss()
    b._C.retval = "helper"
    helper = b.input_code()
    assert events == [("c.input_code", )]
    assert helper == "helper"
    with pytest.raises(errors.OnlyOneCodeError):
        b.input_code()


def test_allocate_code_boss():
    b, events = build_boss()
    wl = object()
    with mock.patch("wormhole._boss.PGPWordList", return_value=wl):
        b.allocate_code(3)
    assert events == [("c.allocate_code", 3, wl)]
    with pytest.raises(errors.OnlyOneCodeError):
        b.allocate_code(3)


def build_rendezvous():
    events = []
    reactor = object()
    journal = ImmediateJournal()
    tor_manager = None
    client_version = ("python", __version__)
    rc = _rendezvous.RendezvousConnector(
        "ws://host:4000/v1", "appid", "side", reactor, journal,
        tor_manager, timing.DebugTiming(), client_version, lambda **kw: None)
    b = Dummy("b", events, IBoss, "error")
    n = Dummy("n", events, INameplate, "connected", "lost")
    m = Dummy("m", events, IMailbox, "connected", "lost")
    a = Dummy("a", events, IAllocator, "connected", "lost")
    x = Dummy("l", events, ILister, "connected", "lost")
    t = Dummy("t", events, ITerminator)
    rc.wire(b, n, m, a, x, t)
    return rc, events


def test_rendezvous_basic():
    rc, events = build_rendezvous()
    del rc, events

def test_websocket_failure():
    # if the TCP connection succeeds, but the subsequent WebSocket
    # negotiation fails, then we'll see an onClose without first seeing
    # onOpen
    rc, events = build_rendezvous()
    rc.ws_close(False, 1006, "connection was closed uncleanly")
    # this should cause the ClientService to be shut down, and an error
    # delivered to the Boss
    assert len(events) == 1, events
    assert events[0][0] == "b.error"
    assert isinstance(events[0][1], errors.ServerConnectionError)
    assert str(events[0][1]) == "connection was closed uncleanly"

def test_websocket_lost():
    # if the TCP connection succeeds, and negotiation completes, then the
    # connection is lost, several machines should be notified
    rc, events = build_rendezvous()

    ws = mock.Mock()

    def notrandom(length):
        return b"\x00" * length

    with mock.patch("os.urandom", notrandom):
        rc.ws_open(ws)
    assert events == [
        ("n.connected", ),
        ("m.connected", ),
        ("l.connected", ),
        ("a.connected", ),
    ]
    events[:] = []

    def sent_messages(ws):
        for c in ws.mock_calls:
            assert c[0] == "sendMessage", ws.mock_calls
            assert not c[1][1]
            yield bytes_to_dict(c[1][0])

    assert list(sent_messages(ws)) == [
            dict(
                appid="appid",
                side="side",
                client_version=["python", __version__],
                id="0000",
                type="bind"),
        ]

    rc.ws_close(True, None, None)
    assert events == [
        ("n.lost", ),
        ("m.lost", ),
        ("l.lost", ),
        ("a.lost", ),
    ]

def test_endpoints():
    # parse different URLs and check the tls status of each
    reactor = object()
    journal = ImmediateJournal()
    tor_manager = None
    client_version = ("python", __version__)
    rc = _rendezvous.RendezvousConnector(
        "ws://host:4000/v1", "appid", "side", reactor, journal,
        tor_manager, timing.DebugTiming(), client_version, lambda **kw: None)

    new_ep = object()
    with mock.patch("twisted.internet.endpoints.HostnameEndpoint",
                    return_value=new_ep) as he:
        ep = rc._make_endpoint("ws://host:4000/v1")
    assert he.mock_calls == [mock.call(reactor, "host", 4000)]
    assert ep is new_ep

    new_ep = object()
    with mock.patch("twisted.internet.endpoints.HostnameEndpoint",
                    return_value=new_ep) as he:
        ep = rc._make_endpoint("ws://host/v1")
    assert he.mock_calls == [mock.call(reactor, "host", 80)]
    assert ep is new_ep

    new_ep = object()
    with mock.patch("twisted.internet.endpoints.clientFromString",
                    return_value=new_ep) as cfs:
        ep = rc._make_endpoint("wss://host:4000/v1")
    assert cfs.mock_calls == [mock.call(reactor, "tls:host:4000")]
    assert ep is new_ep

    new_ep = object()
    with mock.patch("twisted.internet.endpoints.clientFromString",
                    return_value=new_ep) as cfs:
        ep = rc._make_endpoint("wss://host/v1")
    assert cfs.mock_calls == [mock.call(reactor, "tls:host:443")]
    assert ep is new_ep

    tor_manager = mock.Mock()
    directlyProvides(tor_manager, ITorManager)
    rc = _rendezvous.RendezvousConnector(
        "ws://host:4000/v1", "appid", "side", reactor, journal,
        tor_manager, timing.DebugTiming(), client_version, lambda **kw: None)

    tor_manager.mock_calls[:] = []
    ep = rc._make_endpoint("ws://host:4000/v1")
    assert tor_manager.mock_calls == \
                     [mock.call.stream_via("host", 4000, tls=False)]

    tor_manager.mock_calls[:] = []
    ep = rc._make_endpoint("ws://host/v1")
    assert tor_manager.mock_calls == \
                     [mock.call.stream_via("host", 80, tls=False)]

    tor_manager.mock_calls[:] = []
    ep = rc._make_endpoint("wss://host:4000/v1")
    assert tor_manager.mock_calls == \
                     [mock.call.stream_via("host", 4000, tls=True)]

    tor_manager.mock_calls[:] = []
    ep = rc._make_endpoint("wss://host/v1")
    assert tor_manager.mock_calls == \
                     [mock.call.stream_via("host", 443, tls=True)]


# TODO
# #Send
# #Mailbox
# #Nameplate
# #Terminator
# Boss
# RendezvousConnector (not a state machine)
# #Input: exercise helper methods
# #wordlist
# test idempotency / at-most-once where applicable



NOISE_MAX_PAYLOAD = (2**16 - 1) - 16  # 65535 minus 16 bytes authentication data
NOISE_MAX_CIPHERTEXT = (2**16 - 1)  # 65535

try:
    from noise.exceptions import NoiseInvalidMessage
except ImportError:
    class NoiseInvalidMessage(Exception):
        pass

try:
    from noise.exceptions import NoiseHandshakeError
except ImportError:
    class NoiseHandshakeError(Exception):
        pass

try:
    from noise.connection import NoiseConnection
except ImportError:
    # allow imports to work on py2.7, even if dilation doesn't
    NoiseConnection = None



import os
from collections import deque
from attr import attrs, attrib, evolve, define, field
from attr.validators import instance_of, optional
from automat import MethodicalMachine
from zope.interface import implementer
from twisted.internet.defer import Deferred, inlineCallbacks
from twisted.python import log, failure
from .._interfaces import IDilator, IDilationManager, ISend, ITerminator
from ..util import dict_to_bytes, bytes_to_dict, bytes_to_hexstr, provides
from ..observer import OneShotObserver
from .._key import derive_key
from .subchannel import (_WormholeAddress,
                         SubchannelConnectorEndpoint,
                         SubchannelDemultiplex,
                         SubchannelListenerEndpoint)
from .connector import Connector
from .._hints import parse_hint
from .roles import LEADER, FOLLOWER
from .connection import KCM, Ping, Pong, Open, Data, Close, Ack
from .inbound import Inbound
from .outbound import Outbound
from .._status import (DilationStatus, WormholeStatus,
                       ConnectedPeer, ConnectingPeer, ReconnectingPeer, StoppedPeer,
                       )


# exported to Wormhole() for inclusion in versions message
# note that these are strings, not numbers, to facilitate
# experimentation or non-standard versions; the _order_ of versions in
# "can-dilate" is important!
# versions shall be named after wizards from the "Earthsea" series by le Guin
DILATION_VERSIONS = ["ged"]


class OldPeerCannotDilateError(Exception):
    pass


class UnknownDilationMessageType(Exception):
    pass


class ReceivedHintsTooEarly(Exception):
    pass


class UnexpectedKCM(Exception):
    pass


class UnknownMessageType(Exception):
    pass


@define
class DilatedWormhole:
    """
    Represents actions available once a wormhole has been successfully dilated.

    New subchannels to the other peer may be established by first
    obtaining an `IStreamClientEndpoint` from the
    `subprotocol_connector_for("subproto-name")` method. Note that
    ``.connect()`` on these endpoints will ``.errback()`` if Dilation
    cannot be established.
    """

    _manager: IDilationManager = field()

    @inlineCallbacks
    def when_dilated(self):
        yield self._manager._main_channel.when_fired()
        return None

    def listener_for(self, subprotocol_name):
        """
        :returns: an IStreamServerEndpoint that may be used to listen for
           the creation of new subchannels with a particular name.

        Once ``.listen()`` is called on the returned endpoint, every
        new subchannel with this name will have ``.buildProtocol()``
        called, that is what you'd expect Twisted to do.

        (Can we errback something here if we entirely failed to dilate?)
        --> probably only if we make this API async?
        """
        return SubchannelListenerEndpoint(
            subprotocol_name,
            self._manager,
        )

    def connector_for(self, subprotocol_name):
        """
        :returns: an IStreamClientEndpoint that may be used to create new
            subchannels using a specific kind of subprotocol

        Once ``.connect()`` is called on the returned endpoint, a new
        subchannel is opened from this peer to the other peer. The
        other peer sees an OPEN and instantiates a listener from the
        Factory it was given during creation of the wormhole.
        """
        return SubchannelConnectorEndpoint(
            subprotocol_name,
            self._manager,
            self._manager._host_addr,
            self._manager._eventual_queue,
        )


@attrs
class Once:
    _errtype = attrib()

    def __attrs_post_init__(self):
        self._called = False

    def __call__(self):
        if self._called:
            raise self._errtype()
        self._called = True


class CanOnlyDilateOnceError(Exception):
    pass


def make_side():
    return bytes_to_hexstr(os.urandom(8))


# new scheme:
# * both sides send PLEASE as soon as they have an unverified key and
#    w.dilate has been called,
# * PLEASE includes a dilation-specific "side" (independent of the "side"
#    used by mailbox messages)
# * higher "side" is Leader, lower is Follower
# * PLEASE includes the selection of a version from the "can-dilate" list of versions, requires overlap
#    "1" is current

# * we start dilation after both w.dilate() and receiving VERSION, putting us
#   in WANTING, then we process all previously-queued inbound DILATE-n
#   messages. When PLEASE arrives, we move to CONNECTING
# * HINTS sent after dilation starts
# * only Leader sends RECONNECT, only Follower sends RECONNECTING. This
#    is the only difference between the two sides, and is not enforced
#    by the protocol (i.e. if the Follower sends RECONNECT to the Leader,
#    the Leader will obey, although TODO how confusing will this get?)
# * upon receiving RECONNECT: drop Connector, start new Connector, send
#   RECONNECTING, start sending HINTS
# * upon sending RECONNECT: go into FLUSHING state and ignore all HINTS until
#   RECONNECTING received. The new Connector can be spun up earlier, and it
#   can send HINTS, but it must not be given any HINTS that arrive before
#   RECONNECTING (since they're probably stale)

# * after VERSIONS(KCM) received, we might learn that the other side cannot
#    dilate. w.dilate errbacks at this point

# * maybe signal warning if we stay in a "want" state for too long
# * nobody sends HINTS until they're ready to receive
# * nobody sends HINTS unless they've called w.dilate() and received PLEASE
# * nobody connects to inbound hints unless they've called w.dilate()
# * if leader calls w.dilate() but not follower, leader waits forever in
#   "want" (doesn't send anything)
# * if follower calls w.dilate() but not leader, follower waits forever
#   in "want", leader waits forever in "wanted"


def _find_shared_versions(my_versions, their_versions): # -> Option[list]:
    """
    Decide on a best version given a ranked list of our and their
    versions (consisting of arbitrary strings). We prefer a higher
    version from 'our' list over the other list.
    """
    their_dilation_versions = set(their_versions)
    shared_versions = set(my_versions).intersection(their_dilation_versions)
    best_version = None

    if shared_versions:
        # the "best" one is whichever version is highest up the
        # list of acceptable versions
        best = sorted([
            (my_versions.index(v), v)
            for v in shared_versions
        ])
        best_version = best[0][1]

    # dilation_version is the best mutually-compatible version we have
    # with the peer, or None if we have nothing in common
    return best_version


@attrs(eq=False)
class TrafficTimer:
    """
    Tracks when timers have expired versus when traffic (usually
    Pongs) has been seen.

    Will trigger a re-connect (if two timer-intervals expire before we
    see traffic).

    The actual timer (and its length) is controlled by the Manager, as
    is the re-connection logic.
    """

    on_reconnect = attrib()  # a callback when a re-connection attempt is required
    start_timer = attrib()  # a callable that should start the interval timer

    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace", lambda self, f: None)  # pragma: no cover

    @m.state(initial=True)
    def no_connection(self):
        """
        We aren't even connected yet
        """

    @m.state()
    def connected(self):
        """
        We are connected, and have recently seen traffic
        """

    @m.state()
    def idle_traffic(self):
        """
        We haven't seen any data for an interval
        """

    @m.input()
    def interval_elapsed(self):
        """
        One interval is an arbitrary amount of time; when two have
        elapsed, we emit a 'reconnect' signal
        """

    @m.input()
    def traffic_seen(self):
        """
        We have seen some traffic
        """

    @m.input()
    def got_connection(self):
        """
        A connection has been established
        """

    @m.input()
    def lost_connection(self):
        """
        The connection has been lost
        """

    @m.output()
    def signal_reconnect(self):
        self.on_reconnect()

    @m.output()
    def begin_timing(self):
        self.start_timer()

    no_connection.upon(
        got_connection,
        enter=connected,
        outputs=[begin_timing]
    )
    no_connection.upon(
        interval_elapsed,
        enter=no_connection,
        outputs=[begin_timing]
    )
    connected.upon(
        lost_connection,
        enter=no_connection,
        outputs=[]
    )

    connected.upon(
        interval_elapsed,
        enter=idle_traffic,
        outputs=[begin_timing]
    )
    connected.upon(
        traffic_seen,
        enter=connected,
        outputs=[begin_timing]
    )

    idle_traffic.upon(
        interval_elapsed,
        enter=connected,
        outputs=[signal_reconnect]
    )
    idle_traffic.upon(
        traffic_seen,
        enter=connected,
        outputs=[]
    )
    idle_traffic.upon(
        lost_connection,
        enter=no_connection,
        outputs=[]
    )


@attrs(eq=False)
@implementer(IDilationManager)
class Manager:
    _S = attrib(validator=provides(ISend), repr=False)
    _my_side = attrib(validator=instance_of(str))
    _transit_relay_location = attrib(validator=optional(instance_of(str)))
    _reactor = attrib(repr=False)
    _eventual_queue = attrib(repr=False)
    _cooperator = attrib(repr=False)
    _acceptable_versions = attrib()
    _ping_interval = attrib(validator=instance_of(float))
    _expected_subprotocols = attrib()
    # TODO: can this validator work when the parameter is optional?
    _no_listen = attrib(validator=instance_of(bool), default=False)
    _status = attrib(default=None)  # callable([DilationStatus])
    _initial_mailbox_status = attrib(default=None)  # WormholeStatus

    _dilation_key = None
    _tor = None  # TODO
    _timing = None  # TODO
    _next_subchannel_id = None  # initialized in choose_role
    _dilation_version = None  # initialized in got_wormhole_versions
    _main_channel = None  # initialized in __attrs_port_init__
    _subprotocol_factories = None  # initialized in __attrs_port_init__

    m = MethodicalMachine()
    set_trace = getattr(m, "_setTrace", lambda self, f: None)  # pragma: no cover

    def __attrs_post_init__(self):
        self._got_versions_d = Deferred()

        self._my_role = None  # determined upon rx_PLEASE
        self._host_addr = _WormholeAddress()

        self._connection = None
        self._made_first_connection = False
        self._stopped = OneShotObserver(self._eventual_queue)
        self._debug_stall_connector = False

        self._next_dilation_generation = 0
        self._latest_status = DilationStatus(mailbox=self._initial_mailbox_status or WormholeStatus(), generation=0)
        # do not "del" this, the attrs __repr__ gets sad
        self._initial_mailbox_status = None

        # I kept getting confused about which methods were for inbound data
        # (and thus flow-control methods go "out") and which were for
        # outbound data (with flow-control going "in"), so I split them up
        # into separate pieces.
        self._inbound = Inbound(self, self._host_addr)
        self._outbound = Outbound(self, self._cooperator)  # from us to peer

        # TODO: let inbound/outbound create the endpoints, then return them
        # to us
        self._main_channel = OneShotObserver(self._eventual_queue)
        self._subprotocol_factories = SubchannelDemultiplex()

        # NOTE: circular refs, not ideal
        self._api = DilatedWormhole(self)

        # maps outstanding ping_id's (4 bytes) to a 2-tuple (callback, timestamp)
        # (the callback is provided when send_ping is called)
        self._pings_outstanding = dict()

        # Manage our notion of "we have seen traffic recently" or not
        # Only the Leader does this (as only it can decide "we need a
        # new generation")
        self._traffic = None
        self._timer = None

    def _signal_reconnect(self):
        """
        Called by the TrafficTimer machine if we should re-connect (due to
        missed pings)
        """
        if self._connection:
            self._connection.disconnect()

    def _send_ping_reset_timer(self):
        """
        Called by the TrafficTimer machine whenever we should start (or
        extend) our timer and send a ping
        """

        def got_pong(_):
            # ignoring "ping_id"
            self._traffic.traffic_seen()

        self.send_ping(os.urandom(4), got_pong)
        if self._timer is None:
            def timer_expired():
                self._timer = None
                self._traffic.interval_elapsed()
            self._timer = self._reactor.callLater(self._ping_interval, timer_expired)
        else:
            # we already have a timer runner, so extend it
            self._timer.delay(self._ping_interval)

    def _register_subprotocol_factory(self, name, factory):
        """
        Internal helper. Application code has asked to listen for a
        particular subprotocol.  It is an error to listen twice on the
        same subprotocol.
        """
        self._subprotocol_factories.register(name, factory)

    def got_dilation_key(self, key):
        assert isinstance(key, bytes)
        self._dilation_key = key

    def got_wormhole_versions(self, their_wormhole_versions):
        # this always happens before received_dilation_message
        self._dilation_version = _find_shared_versions(
            self._acceptable_versions,
            their_wormhole_versions.get("can-dilate", [])
        )

        if not self._dilation_version:  # "ged" or None
            # TODO: be more specific about the error. dilation_version==None
            # means we had no version in common with them, which could either
            # be because they're so old they don't dilate at all, or because
            # they're so new that they no longer accommodate our old version
            self.fail(failure.Failure(OldPeerCannotDilateError()))

        self.start()

    # from _boss.Boss
    def _wormhole_status(self, wormhole_status):
        self._maybe_send_status(
            evolve(
                self._latest_status,
                mailbox=wormhole_status,
            )
        )

    def _maybe_send_status(self, status_msg):
        self._latest_status = status_msg
        if self._status is not None:
            self._status(status_msg)

    def _hint_status(self, hints):
        """
        Internal helper. From Connector, calls to update the hints we're
        actually using
        """
        self._maybe_send_status(
            evolve(
                self._latest_status,
                hints=set(hints).union(self._latest_status.hints),
            )
        )

    def fail(self, f):
        self._main_channel.error(f)

    def received_dilation_message(self, plaintext):
        # this receives new in-order DILATE-n payloads, decrypted but not
        # de-JSONed.

        message = bytes_to_dict(plaintext)
        type = message["type"]
        if type == "please":
            self.rx_PLEASE(message)
        elif type == "connection-hints":
            self.rx_HINTS(message)
            # todo: could be useful to put "hints" in status, and send
            # a status update when getting new hints?
        elif type == "reconnect":
            self.rx_RECONNECT()
        elif type == "reconnecting":
            self.rx_RECONNECTING()
        else:
            log.err(UnknownDilationMessageType(message))
            return

    def when_stopped(self):
        return self._stopped.when_fired()

    def send_dilation_generation(self, **fields):
        dilation_generation = self._next_dilation_generation
        self._next_dilation_generation += 1
        self._S.send("dilate-%d" % dilation_generation, dict_to_bytes(fields))

    def send_hints(self, hints):  # from Connector
        self.send_dilation_generation(type="connection-hints", hints=hints)

    # forward inbound-ish things to _Inbound

    def subchannel_pauseProducing(self, sc):
        self._inbound.subchannel_pauseProducing(sc)

    def subchannel_resumeProducing(self, sc):
        self._inbound.subchannel_resumeProducing(sc)

    def subchannel_stopProducing(self, sc):
        self._inbound.subchannel_stopProducing(sc)

    def subchannel_local_open(self, scid, sc):
        self._inbound.subchannel_local_open(scid, sc)

    # forward outbound-ish things to _Outbound
    def subchannel_registerProducer(self, sc, producer, streaming):
        self._outbound.subchannel_registerProducer(sc, producer, streaming)

    def subchannel_unregisterProducer(self, sc):
        self._outbound.subchannel_unregisterProducer(sc)

    def send_open(self, scid, subprotocol):
        assert isinstance(scid, int)
        self._queue_and_send(Open, scid, subprotocol)

    def send_data(self, scid, data):
        assert isinstance(scid, int)
        self._queue_and_send(Data, scid, data)

    def send_close(self, scid):
        assert isinstance(scid, int)
        self._queue_and_send(Close, scid)

    def _queue_and_send(self, record_type, *args):
        r = self._outbound.build_record(record_type, *args)
        # Outbound owns the send_record() pipe, so that it can stall new
        # writes after a new connection is made until after all queued
        # messages are written (to preserve ordering).
        self._outbound.queue_and_send_record(r)  # may trigger pauseProducing

    def subchannel_closed(self, scid, sc):
        # let everyone clean up. This happens just after we delivered
        # connectionLost to the Protocol, except for the control channel,
        # which might get connectionLost later after they use ep.connect.
        # TODO: is this inversion a problem?
        self._inbound.subchannel_closed(scid, sc)
        self._outbound.subchannel_closed(scid, sc)

    # our Connector calls these

    def connector_connection_made(self, c):
        # only the "Leader" will do pings, because it is the only side
        # which can decide "our peer connection is dead"
        if self._my_role == LEADER:
            # if we have just RE-connected, then we'll already have a
            # _traffic instance but the first time we connect we do not
            if self._traffic is None:
                self._traffic = TrafficTimer(self._signal_reconnect, self._send_ping_reset_timer)
            self._traffic.got_connection()

        self.connection_made()  # state machine update
        self._connection = c
        self._inbound.use_connection(c)
        self._outbound.use_connection(c)  # does c.registerProducer
        if not self._made_first_connection:
            self._made_first_connection = True
            # might be ideal to send information about our selected
            # Peer connection through here
            self._main_channel.fire(None)
        pass

    def connector_connection_lost(self):
        # ultimately called after a DilatedConnectionProtocol disconnects
        if self._traffic is not None:
            self._traffic.lost_connection()
        self._stop_using_connection()
        if self._my_role is LEADER:
            self.connection_lost_leader()  # state machine
        else:
            self.connection_lost_follower()

    def _stop_using_connection(self):
        # the connection is already lost by this point
        if self._timer is not None:
            self._timer.cancel()
            self._timer = None
        self._connection = None
        self._inbound.stop_using_connection()
        self._outbound.stop_using_connection()  # does c.unregisterProducer

    # from our active Connection

    def got_record(self, r):
        # records with sequence numbers: always ack, ignore old ones
        if isinstance(r, (Open, Data, Close)):
            self.send_ack(r.seqnum)  # always ack, even for old ones
            if self._inbound.is_record_old(r):
                return
            self._inbound.update_ack_watermark(r.seqnum)
            if isinstance(r, Open):
                self._inbound.handle_open(r.scid, r.subprotocol)
            elif isinstance(r, Data):
                self._inbound.handle_data(r.scid, r.data)
            else:  # isinstance(r, Close)
                self._inbound.handle_close(r.scid)
            return
        if isinstance(r, KCM):
            log.err(UnexpectedKCM())
        elif isinstance(r, Ping):
            self.handle_ping(r.ping_id)
        elif isinstance(r, Pong):
            self.handle_pong(r.ping_id)
        elif isinstance(r, Ack):
            self._outbound.handle_ack(r.resp_seqnum)  # retire queued messages
        else:
            log.err(UnknownMessageType(f"{r}"))
        # todo: it might be better to tell the TrafficTimer
        # state-machine every time we see _any_ traffic (i.e. here)
        # -- currently we're demanding that we see the "Pong"
        # if self._traffic is not None:
        #     self._traffic.traffic_seen()

    # pings, pongs, and acks are not queued
    def send_ping(self, ping_id, on_pong=None):
        # ping_id is 4 bytes
        assert ping_id not in self._pings_outstanding, "Duplicate ping_id"
        self._pings_outstanding[ping_id] = (on_pong, self._reactor.seconds())
        self._outbound.send_if_connected(Ping(ping_id))

    def send_pong(self, ping_id):  # ping_id is bytes?
        self._outbound.send_if_connected(Pong(ping_id))

    def send_ack(self, resp_seqnum):
        self._outbound.send_if_connected(Ack(resp_seqnum))

    def handle_ping(self, ping_id):
        self._peer_saw_ping()
        self.send_pong(ping_id)

    def handle_pong(self, ping_id):
        if ping_id not in self._pings_outstanding:
            print("Weird: pong for ping that isn't outstanding")
        else:
            self._peer_saw_ping()
            on_pong, start = self._pings_outstanding.pop(ping_id)
            if on_pong is not None:
                on_pong(self._reactor.seconds() - start)
        # TODO: update is-alive timer

    # status

    def have_peer(self, conn):
        """
        Signal that we have selected a peer connection to use
        """
        self._maybe_send_status(
            evolve(
                self._latest_status,
                peer_connection=ConnectedPeer(
                    self._reactor.seconds(),
                    self._reactor.seconds() + (self._ping_interval * 2),
                    conn._description,
                ),
            )
        )

    def _peer_saw_ping(self):
        """
        We have just seen Ping or Pong traffic from our peer.

        Note that only the Leader sends Pings so one side will see
        only Pongs and one will see only Pings.
        """
        if isinstance(self._latest_status.peer_connection, ConnectedPeer):
            self._maybe_send_status(
                evolve(
                    self._latest_status,
                    peer_connection=evolve(
                        self._latest_status.peer_connection,
                        expires_at=self._reactor.seconds() + (self._ping_interval * 2),
                    )
                )
            )

    # subchannel maintenance
    def allocate_subchannel_id(self):
        scid_num = self._next_subchannel_id
        self._next_subchannel_id += 2
        return scid_num

    # state machine

    @m.state(initial=True)
    def WAITING(self):
        pass  # pragma: no cover

    @m.state()
    def WANTING(self):
        pass  # pragma: no cover

    @m.state()
    def CONNECTING(self):
        pass  # pragma: no cover

    @m.state()
    def CONNECTED(self):
        pass  # pragma: no cover

    @m.state()
    def FLUSHING(self):
        pass  # pragma: no cover

    @m.state()
    def ABANDONING(self):
        pass  # pragma: no cover

    @m.state()
    def LONELY(self):
        pass  # pragma: no cover

    @m.state()
    def STOPPING(self):
        pass  # pragma: no cover

    @m.state(terminal=True)
    def STOPPED(self):
        pass  # pragma: no cover

    @m.input()
    def start(self):
        pass  # pragma: no cover

    @m.input()
    def rx_PLEASE(self, message):
        pass  # pragma: no cover

    @m.input()  # only sent by Follower
    def rx_HINTS(self, hint_message):
        pass  # pragma: no cover

    @m.input()  # only Leader sends RECONNECT, so only Follower receives it
    def rx_RECONNECT(self):
        pass  # pragma: no cover

    @m.input()  # only Follower sends RECONNECTING, so only Leader receives it
    def rx_RECONNECTING(self):
        pass  # pragma: no cover

    # Connector gives us connection_made()
    @m.input()
    def connection_made(self):
        pass  # pragma: no cover

    # our connection_lost() fires connection_lost_leader or
    # connection_lost_follower depending upon our role. If either side sees a
    # problem with the connection (timeouts, bad authentication) then they
    # just drop it and let connection_lost() handle the cleanup.
    @m.input()
    def connection_lost_leader(self):
        pass  # pragma: no cover

    @m.input()
    def connection_lost_follower(self):
        pass

    @m.input()
    def stop(self):
        pass  # pragma: no cover

    @m.output()
    def send_please(self):
        msg = {
            "type": "please",
            "side": self._my_side,
        }
        if self._dilation_version is not None:
            msg["use-version"] = self._dilation_version
        self.send_dilation_generation(**msg)

    @m.output()
    def choose_role(self, message):
        their_side = message["side"]
        if self._my_side > their_side:
            self._my_role = LEADER
            # scid 0 is reserved for the control channel. the leader uses odd
            # numbers starting with 1
            self._next_subchannel_id = 1
        elif their_side > self._my_side:
            self._my_role = FOLLOWER
            # the follower uses even numbers starting with 2
            self._next_subchannel_id = 2
        else:
            raise ValueError("their side shouldn't be equal: reflection?")

    # these Outputs behave differently for the Leader vs the Follower

    @m.output()
    def start_connecting_ignore_message(self, message):
        del message  # ignored
        return self._start_connecting()

    @m.output()
    def start_connecting(self):
        self._start_connecting()

    def _start_connecting(self):
        assert self._my_role is not None
        assert self._dilation_key is not None
        self._connector = Connector(self._dilation_key,
                                    self._transit_relay_location,
                                    self,
                                    self._reactor, self._eventual_queue,
                                    self._no_listen, self._tor,
                                    self._timing,
                                    self._my_side,  # needed for relay handshake
                                    self._my_role)
        if self._debug_stall_connector:
            # unit tests use this hook to send messages while we know we
            # don't have a connection
            self._eventual_queue.eventually(self._debug_stall_connector, self._connector)
            return
        self._connector.start()

    @m.output()
    def send_reconnect(self):
        self.send_dilation_generation(type="reconnect")  # TODO: generation number?

    @m.output()
    def send_reconnecting(self):
        self.send_dilation_generation(type="reconnecting")  # TODO: generation?

    @m.output()
    def use_hints(self, hint_message):
        hint_objs = filter(lambda h: h,  # ignore None, unrecognizable
                           [parse_hint(hs) for hs in hint_message["hints"]])
        hint_objs = list(hint_objs)
        self._connector.got_hints(hint_objs)

    @m.output()
    def stop_connecting(self):
        self._connector.stop()

    @m.output()
    def abandon_connection(self):
        # we think we're still connected, but the Leader disagrees. Or we've
        # been told to shut down.
        if self._timer is not None:
            self._timer.cancel()
            self._timer = None
        self._connection.disconnect()  # let connection_lost do cleanup

    @m.output()
    def notify_stopped(self):
        self._stopped.fire(None)

    @m.output()
    def send_status_connecting(self):
        self._maybe_send_status(
            evolve(
                self._latest_status,
                peer_connection=ConnectingPeer(self._reactor.seconds()),
            )
        )

    @m.output()
    def send_status_reconnecting(self):
        self._maybe_send_status(
            evolve(
                self._latest_status,
                peer_connection=ReconnectingPeer(self._reactor.seconds()),
            )
        )

    @m.output()
    def send_status_dilation_generation(self):
        # send_dilation_generation has just run recently, incrementing
        # this; "current status" is thus the prior value
        dilation_generation = self._next_dilation_generation - 1
        self._maybe_send_status(
            evolve(
                self._latest_status,
                generation=dilation_generation,
            )
        )

    @m.output()
    def send_status_stopped(self):
        self._maybe_send_status(
            evolve(
                self._latest_status,
                peer_connection=StoppedPeer(),
            )
        )

    # We are born WAITING after the local app calls w.dilate(). We enter
    # WANTING (and send a PLEASE) when we learn of a mutually-compatible
    # dilation_version.
    WAITING.upon(start, enter=WANTING, outputs=[send_please, send_status_dilation_generation])

    # we start CONNECTING when we get rx_PLEASE
    WANTING.upon(rx_PLEASE, enter=CONNECTING,
                 outputs=[choose_role, start_connecting_ignore_message, send_status_connecting])

    CONNECTING.upon(connection_made, enter=CONNECTED, outputs=[])

    # Leader
    CONNECTED.upon(connection_lost_leader, enter=FLUSHING,
                   outputs=[send_reconnect, send_status_dilation_generation, send_status_reconnecting])
    FLUSHING.upon(rx_RECONNECTING, enter=CONNECTING,
                  outputs=[start_connecting, send_status_reconnecting])

    # Follower
    # if we notice a lost connection, just wait for the Leader to notice too
    CONNECTED.upon(connection_lost_follower, enter=LONELY, outputs=[])
    LONELY.upon(rx_RECONNECT, enter=CONNECTING,
                outputs=[send_reconnecting, start_connecting,
                         send_status_dilation_generation, send_status_reconnecting])
    # but if they notice it first, abandon our (seemingly functional)
    # connection, then tell them that we're ready to try again
    CONNECTED.upon(rx_RECONNECT, enter=ABANDONING, outputs=[abandon_connection])
    ABANDONING.upon(connection_lost_follower, enter=CONNECTING,
                    outputs=[send_reconnecting, start_connecting,
                             send_status_dilation_generation, send_status_reconnecting])
    # and if they notice a problem while we're still connecting, abandon our
    # incomplete attempt and try again. in this case we don't have to wait
    # for a connection to finish shutdown
    CONNECTING.upon(rx_RECONNECT, enter=CONNECTING,
                    outputs=[stop_connecting,
                             send_reconnecting,
                             start_connecting,
                             send_status_dilation_generation,
                             send_status_reconnecting])

    # rx_HINTS never changes state, they're just accepted or ignored
    WANTING.upon(rx_HINTS, enter=WANTING, outputs=[])  # too early
    CONNECTING.upon(rx_HINTS, enter=CONNECTING, outputs=[use_hints])
    CONNECTED.upon(rx_HINTS, enter=CONNECTED, outputs=[])  # too late, ignore
    FLUSHING.upon(rx_HINTS, enter=FLUSHING, outputs=[])  # stale, ignore
    LONELY.upon(rx_HINTS, enter=LONELY, outputs=[])  # stale, ignore
    ABANDONING.upon(rx_HINTS, enter=ABANDONING, outputs=[])  # shouldn't happen
    STOPPING.upon(rx_HINTS, enter=STOPPING, outputs=[])

    WAITING.upon(stop, enter=STOPPED, outputs=[notify_stopped])
    WANTING.upon(stop, enter=STOPPED, outputs=[notify_stopped])
    CONNECTING.upon(stop, enter=STOPPED, outputs=[stop_connecting, notify_stopped, send_status_stopped])
    CONNECTED.upon(stop, enter=STOPPING, outputs=[abandon_connection])
    ABANDONING.upon(stop, enter=STOPPING, outputs=[])
    FLUSHING.upon(stop, enter=STOPPED, outputs=[notify_stopped, send_status_stopped])
    LONELY.upon(stop, enter=STOPPED, outputs=[notify_stopped, send_status_stopped])
    STOPPING.upon(connection_lost_leader, enter=STOPPED, outputs=[notify_stopped, send_status_stopped])
    STOPPING.upon(connection_lost_follower, enter=STOPPED, outputs=[notify_stopped, send_status_stopped])


@attrs
@implementer(IDilator)
class Dilator:
    """I launch the dilation process.

    I am created with every Wormhole (regardless of whether .dilate()
    was called or not), and I handle the initial phase of dilation,
    before we know whether we'll be the Leader or the Follower. Once we
    hear the other side's VERSION message (which tells us that we have a
    connection, they are capable of dilating, and which side we're on),
    then we build a Manager and hand control to it.
    """

    _reactor = attrib()
    _eventual_queue = attrib()
    _cooperator = attrib()
    _acceptable_versions = attrib()
    _did_dilate = attrib(init=False)

    def __attrs_post_init__(self):
        self._manager = None
        self._pending_dilation_key = None
        self._pending_wormhole_versions = None
        self._pending_inbound_dilate_messages = deque()
        self._did_dilate = Once(CanOnlyDilateOnceError)

    def wire(self, sender, terminator):
        self._S = ISend(sender)
        self._T = ITerminator(terminator)

    # this is the primary entry point, called when w.dilate() is
    # invoked; upstream calls are basically just call-through -- so
    # all these inputs should be validated.
    def dilate(self, transit_relay_location=None, no_listen=False, wormhole_status=None, status_update=None,
               ping_interval=None, expected_subprotocols=None):
        # ensure users can only call this API once -- in the past, it
        # was possible to call the API more than once but any cal
        # after the first would have no real effect:
        # transit_relay_location, no_listen, etc would all remain
        # unchanged)
        self._did_dilate()

        if self._manager is None:
            # build the manager right away, and tell it later when the
            # VERSIONS message arrives, and also when the dilation_key is set
            my_dilation_side = make_side()
            m = Manager(
                self._S,
                my_dilation_side,
                transit_relay_location,
                self._reactor,
                self._eventual_queue,
                self._cooperator,
                self._acceptable_versions,
                ping_interval or 30.0,
                expected_subprotocols,
                no_listen,
                status_update,
                initial_mailbox_status=wormhole_status,
            )
            self._manager = m
            if self._pending_dilation_key is not None:
                m.got_dilation_key(self._pending_dilation_key)
            if self._pending_wormhole_versions:
                m.got_wormhole_versions(self._pending_wormhole_versions)
            while self._pending_inbound_dilate_messages:
                plaintext = self._pending_inbound_dilate_messages.popleft()
                m.received_dilation_message(plaintext)

        return self._manager._api

    # Called by Terminator after everything else (mailbox, nameplate, server
    # connection) has shut down. Expects to fire T.stoppedD() when Dilator is
    # stopped too.
    def stop(self):
        if self._manager:
            self._manager.stop()
            # TODO: avoid Deferreds for control flow, hard to serialize
            self._manager.when_stopped().addCallback(lambda _: self._T.stoppedD())
        else:
            self._T.stoppedD()
            return
        # TODO: tolerate multiple calls

    # from Boss

    def got_key(self, key):
        # TODO: verify this happens before got_wormhole_versions, or add a gate
        # to tolerate either ordering
        purpose = b"dilation-v1"
        LENGTH = 32  # TODO: whatever Noise wants, I guess
        dilation_key = derive_key(key, purpose, LENGTH)
        if self._manager:
            self._manager.got_dilation_key(dilation_key)
        else:
            self._pending_dilation_key = dilation_key

    def got_wormhole_versions(self, their_wormhole_versions):
        if self._manager:
            self._manager.got_wormhole_versions(their_wormhole_versions)
        else:
            self._pending_wormhole_versions = their_wormhole_versions

    def received_dilate(self, plaintext):
        if not self._manager:
            self._pending_inbound_dilate_messages.append(plaintext)
        else:
            self._manager.received_dilation_message(plaintext)
